# Auto generated from test.yaml by pythongen.py version: 0.9.0
# Generation date: 2023-08-17T14:01:39
# Schema: microscopemetrics_samples_field_illumination_schema
#
# id: https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml
# description:
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import Float

metamodel_version = "1.7.0"
version = None

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
DEFAULT_ = CurieNamespace('', 'https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/')


# Types

# Class references



@dataclass
class MyClass(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef("https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/MyClass")
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "MyClass"
    class_model_uri: ClassVar[URIRef] = URIRef("https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/MyClass")

    my_slot: Optional[Union[float, List[float]]] = 0.3

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.my_slot, list):
            self.my_slot = [self.my_slot] if self.my_slot is not None else []
        self.my_slot = [v if isinstance(v, float) else float(v) for v in self.my_slot]

        super().__post_init__(**kwargs)


# Enumerations


# Slots
class slots:
    pass

slots.my_slot = Slot(uri=DEFAULT_.my_slot, name="my_slot", curie=DEFAULT_.curie('my_slot'),
                   model_uri=DEFAULT_.my_slot, domain=None, range=Optional[Union[float, List[float]]])



from datetime import datetime
from typing import Dict, Tuple

import numpy as np
import scipy
from numpy import float64, ndarray
from skimage.filters import gaussian
from skimage.measure import regionprops

import microscopemetrics.data_schema.samples.field_illumination_schema as schema
from microscopemetrics.samples import AnalysisMixin, logger, numpy_to_inlined_image
from microscopemetrics.utilities.utilities import is_saturated


def _channel_intensity_map(channel, map_size: int):
    """
    Compute the intensity map of a channel
    Parameters
    ----------
    channel : np.array.
        image on a 2d ndarray format.
    map_size : int
        size of the intensity map.
    Returns
    -------
    intensity_map : ndarray
        2d ndarray representing the intensity map of the chosen channel.
    """
    channel = channel / channel.max()
    return scipy.ndimage.zoom(channel, map_size / channel.shape[0])


def _image_intensity_map(image: np.ndarray, map_size: int):
    """
    Compute the intensity map of an image
    Parameters
    ----------
    image : ndarray.
        image on a 3d ndarray format yxc.
    map_size : int
        size of the intensity map.
    Returns
    -------
    intensity_map : ndarray
        3d ndarray representing the intensity map of the chosen image.
    """
    output = np.zeros((map_size, map_size, image.shape[2]))
    for c in range(image.shape[2]):
        output[:, :, c] = _channel_intensity_map(np.squeeze(image[:, :, c]), map_size)

    # We want to return a 5d array (adding z and t) for compatibility with the rest of the code
    return np.expand_dims(output, axis=(0, 1))


def _channel_line_profile(
    channel: ndarray, start: Tuple[int, int], end: Tuple[int, int], profile_size: int
) -> ndarray:
    """
    Compute the intensity profile along a line between x0-y0 and x1-y1 using cubic interpolation
    Parameters
    ----------
    channel : np.array.
        image on a 2d ndarray format.
    start : (int, int)
        coordinates of the starting pixel
    end : (int, int)
        coordinates of the ending pixel
    Returns
    -------
    line_pixel_values : ndarray
        1d ndarray representing the values of the chosen line of pixels.
    """
    x, y = np.linspace(start[0], end[0], profile_size), np.linspace(start[1], end[1], profile_size)

    return scipy.ndimage.map_coordinates(channel, np.vstack((x, y)))


def _image_line_profile(image: ndarray, profile_size: int):
    """
    Compute the intensity profile along a line between x0-y0 and x1-y1
    Parameters
    ----------
    image : ndarray.
        image on a 3d ndarray format yxc.
    profile_size : int
        size of the intensity profile.
    Returns
    -------
    line_pixel_values : ndarray
        2d ndarray representing the values of the chosen line of pixels for each channel.
    """
    profile_coordinates = {
        "leftTop_to_rightBottom": ((0, 0), (image.shape[1], image.shape[0])),
        "leftBottom_to_rightTop": ((0, image.shape[0]), (image.shape[1], 0)),
        "center_horizontal": (
            (0, image.shape[0] // 2),
            (image.shape[1], image.shape[0] // 2),
        ),
        "center_vertical": (
            (image.shape[1] // 2, 0),
            (image.shape[1] // 2, image.shape[0]),
        ),
    }
    output = []
    for profile_name, (start, end) in profile_coordinates.items():
        profiles = np.zeros((image.shape[2], 255))
        for c in range(image.shape[2]):
            profiles[c, :] = _channel_line_profile(
                np.squeeze(image[:, :, c]), start, end, profile_size
            )
        output = output + [
            {f"ch_{c}_{profile_name}": {"values": profiles[c].tolist()}}
            for c in range(image.shape[2])
        ]

    return output


def _segment_channel(channel, threshold: float, sigma: float):
    if sigma is not None:
        channel = gaussian(image=channel, sigma=sigma, preserve_range=True, channel_axis=None)

    channel_norm = channel / np.max(channel)
    return (channel_norm > threshold).astype(int)


def _channel_max_intensity_properties(
    channel: np.array, sigma: float, center_threshold: float
) -> dict:
    """Computes the center of mass and the max intensity of the maximum intensity region of an image.
    Parameters
    ----------
    channel : np.array.
        2d ndarray.
    Returns
    -------
    center_of_mass: dict
        dict enclosing the number of pixels, the coordinates of the
        center of mass of the and the max intensity value of the max intensity
        area of the provided image.
    """
    max_intensity = np.max(channel)
    # TODO: check if there is more than one pixel with the same intensity.
    #  We take the first one but we should take the one closest to the center or use a find_peaks function
    max_intensity_indexes = np.unravel_index(np.argmax(channel), channel.shape)

    max_int_mask = _segment_channel(channel, center_threshold, sigma)
    image_properties = regionprops(max_int_mask, channel)

    return {
        "nb_pixels": image_properties[0].area,
        "center_of_mass_x": image_properties[0].centroid_weighted[0],
        "center_of_mass_y": image_properties[0].centroid_weighted[1],
        "max_intensity": max_intensity,
        "max_intensity_pos_x": max_intensity_indexes[0],
        "max_intensity_pos_y": max_intensity_indexes[1],
    }


def _channel_corner_properties(channel, corner_fraction=0.1):
    max_intensity = np.max(channel)

    # Calculate the corner fraction in pixels (cfp) of the image size
    # to use as the corner size and the center range (cr)
    cfp = int(corner_fraction * (channel.shape[0] + channel.shape[1]) / 2)
    cr_y = int((channel.shape[0] - cfp) / 2)
    cr_x = int((channel.shape[1] - cfp) / 2)

    return {
        "top_left_intensity_mean": np.mean(channel[0:cfp, 0:cfp]),
        "top_left_intensity_ratio": np.mean(channel[0:cfp, 0:cfp]) / max_intensity,
        "top_center_intensity_mean": np.mean(channel[cr_x:-cr_x, 0:cfp]),
        "top_center_intensity_ratio": np.mean(channel[cr_x:-cr_x, 0:cfp]) / max_intensity,
        "top_right_intensity_mean": np.mean(channel[-cfp:-1, 0:cfp]),
        "top_right_intensity_ratio": np.mean(channel[-cfp:-1, 0:cfp]) / max_intensity,
        "middle_left_intensity_mean": np.mean(channel[0:cfp, cr_y:-cr_y]),
        "middle_left_intensity_ratio": np.mean(channel[0:cfp, cr_y:-cr_y]) / max_intensity,
        "middle_center_intensity_mean": np.mean(channel[cr_x:-cr_x, cr_y:-cr_y]),
        "middle_center_intensity_ratio": np.mean(channel[cr_x:-cr_x, cr_y:-cr_y]) / max_intensity,
        "middle_right_intensity_mean": np.mean(channel[-cfp:-1, cr_y:-cr_y]),
        "middle_right_intensity_ratio": np.mean(channel[-cfp:-1, cr_y:-cr_y]) / max_intensity,
        "bottom_left_intensity_mean": np.mean(channel[0:cfp, -cfp:-1]),
        "bottom_left_intensity_ratio": np.mean(channel[0:cfp, -cfp:-1]) / max_intensity,
        "bottom_center_intensity_mean": np.mean(channel[cr_x:-cr_x, -cfp:-1]),
        "bottom_center_intensity_ratio": np.mean(channel[cr_x:-cr_x, -cfp:-1]) / max_intensity,
        "bottom_right_intensity_mean": np.mean(channel[-cfp:-1, -cfp:-1]),
        "bottom_right_intensity_ratio": np.mean(channel[-cfp:-1, -cfp:-1]) / max_intensity,
    }


def _channel_area_deciles(channel: ndarray) -> dict:
    """Computes the intensity deciles of an image.
    Parameters
    ----------
    channel : np.array.
        2d ndarray.
    Returns
    -------
    deciles: dict
        dict enclosing the intensity deciles of the provided channel.
    """
    channel = channel / np.max(channel)
    return {f"decile_{i}": np.percentile(channel, i * 10) for i in range(10)}


def _image_properties(
    image: ndarray, corner_fraction: float, sigma: float, center_threshold: float
):
    """
    given an image in a 3d ndarray format (yxc), this function return intensities for the corner and central regions
    and their ratio over the maximum intensity value of the array.
    Parameters
    ----------
    image : ndarray
        image on a 2d np.ndarray in yxc format.
    Returns
    -------
    profiles_statistics : dict
        Dictionary showing the intensity values of the different regions and
        their ratio over the maximum intensity value of the array.
        Dictionary values will be lists in case of multiple channels.
    """
    properties = []
    for c in range(image.shape[2]):
        channel_properties = {"channel": c}
        channel_properties.update(
            _channel_max_intensity_properties(image[:, :, c], sigma, center_threshold)
        )
        channel_properties.update(_channel_corner_properties(image[:, :, c], corner_fraction))
        channel_properties.update(_channel_area_deciles(image[:, :, c]))
        if image.shape[2] == 1:
            return channel_properties
        else:
            properties.append(channel_properties)

    properties = {k: [i[k] for i in properties] for k in properties[0]}
    return properties


class FieldIlluminationAnalysis(schema.FieldIlluminationDataset, AnalysisMixin):
    """This analysis creates a report on field illumination homogeneity based on input images"""

    def run(self) -> bool:
        self.validate_requirements()

        # Check image shape
        logger.info("Checking image shape...")
        image = self.input.field_illumination_image.data
        if len(image.shape) != 5:
            logger.error("Image must be 5D")
            return False
        if image.shape[0] != 1 or image.shape[1] != 1:
            logger.warning(
                "Image must be in TZYXC order, single z and single time-point. Using first z and time-point."
            )
        # For the analysis we are using only the first z and time-point
        image = image[0, 0, :, :, :].reshape((image.shape[2], image.shape[3], image.shape[4]))

        # Check image saturation
        logger.info("Checking image saturation...")
        saturated_channels = []
        for c in range(image.shape[2]):
            if is_saturated(
                channel=image[:, :, c],
                threshold=self.input.saturation_threshold,
                detector_bit_depth=self.input.bit_depth,
            ):
                logger.error(f"Channel {c} is saturated")
                saturated_channels.append(c)
        if len(saturated_channels):
            logger.error(f"Channels {saturated_channels} are saturated")
            return False

        self.output.key_values = schema.FieldIlluminationKeyValues(
            **_image_properties(
                image=image,
                corner_fraction=self.input.corner_fraction,
                sigma=self.input.sigma,
                center_threshold=self.input.center_threshold,
            )
        )

        self.output.intensity_map = numpy_to_inlined_image(
            array=_image_intensity_map(image=image, map_size=self.input.intensity_map_size),
            name=f"{self.input.field_illumination_image.name}_intensity_map",
            description=f"Intensity map of {self.input.field_illumination_image.name}",
            image_url=self.input.field_illumination_image.image_url,
            source_image_url=self.input.field_illumination_image.source_image_url,
        )

        self.output.intensity_profiles = schema.TableAsDict(
            columns=_image_line_profile(image, profile_size=255)
        )

        self.processing_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.processed = True

        return True



# This is a place to hold mere utilities for metrics

import json
import warnings
from configparser import ConfigParser
from typing import Optional

import numpy as np
from numpy import float64, ndarray
from scipy import special

DETECTOR_BIT_DEPTHS = [10, 11, 12, 15]


## Some useful functions
def convert_SI(val, unit_in, unit_out):
    si = {
        "nanometer": 0.000000001,
        "micrometer": 0.000001,
        "millimeter": 0.001,
        "meter": 1.0,
    }
    return val * si[unit_in.lower()] / si[unit_out.lower()]


# def airy_fun(x, centre, a, exp):  # , amp, bg):
#     if (x - centre) == 0:
#         return a * .5 ** exp
#     else:
#         return a * (special.j1(x - centre) / (x - centre)) ** exp
#
#
# def multi_airy_fun(x, *params):
#     y = np.zeros_like(x)
#     for i in range(0, len(params), 3):
#         y = y + airy_fun(x, params[i], params[i+1], params[i+2])
#     return y
def airy_fun(x: ndarray, centre: float64, amp: float64) -> ndarray:  # , exp):  # , amp, bg):
    with np.errstate(divide="ignore", invalid="ignore"):
        return np.where(
            (x - centre) == 0,
            amp * 0.5**2,
            amp * (special.j1(x - centre) / (x - centre)) ** 2,
        )


def gaussian_fun(x, background, amplitude, center, sd):
    gauss = np.exp(-np.power(x - center, 2.0) / (2 * np.power(sd, 2.0)))
    return background + (amplitude - background) * gauss


def multi_airy_fun(x: ndarray, *params) -> ndarray:
    y = np.zeros_like(x)
    for i in range(0, len(params), 2):
        y = y + airy_fun(x, params[i], params[i + 1])
    return y


def wavelength_to_rgb(wavelength, gamma=0.8):

    """
    Copied from https://www.noah.org/wiki/Wavelength_to_RGB_in_Python
    This converts a given wavelength of light to an
    approximate RGB color value. The wavelength must be given
    in nanometers in the range from 380 nm through 750 nm
    (789 THz through 400 THz).

    Based on code by Dan Bruton
    http://www.physics.sfasu.edu/astro/color/spectra.html
    """

    wavelength = float(wavelength)
    if 380 < wavelength <= 440:
        attenuation = 0.3 + 0.7 * (wavelength - 380) / (440 - 380)
        r = ((-(wavelength - 440) / (440 - 380)) * attenuation) ** gamma
        g = 0.0
        b = (1.0 * attenuation) ** gamma
    elif 440 < wavelength <= 490:
        r = 0.0
        g = ((wavelength - 440) / (490 - 440)) ** gamma
        b = 1.0
    elif 490 < wavelength <= 510:
        r = 0.0
        g = 1.0
        b = (-(wavelength - 510) / (510 - 490)) ** gamma
    elif 510 < wavelength <= 580:
        r = ((wavelength - 510) / (580 - 510)) ** gamma
        g = 1.0
        b = 0.0
    elif 580 < wavelength <= 645:
        r = 1.0
        g = (-(wavelength - 645) / (645 - 580)) ** gamma
        b = 0.0
    elif 645 < wavelength <= 750:
        attenuation = 0.3 + 0.7 * (750 - wavelength) / (750 - 645)
        r = (1.0 * attenuation) ** gamma
        g = 0.0
        b = 0.0
    else:
        r = 0.0
        g = 0.0
        b = 0.0
    r *= int(r * 255)
    g *= int(g * 255)
    g *= int(g * 255)
    return r, g, b


# This class is not used after introduction of pydantic.
class MetricsConfig(ConfigParser):
    def getjson(self, section, option, **kwargs):
        value = self.get(section, option, **kwargs)
        try:
            output = json.loads(value)
        except json.JSONDecodeError as e:
            raise e
        return output

    def getlist(self, section, option, **kwargs):
        output = self.getjson(section, option, **kwargs)

        if type(output) is list:
            return output
        else:
            raise Exception(
                f'The config option "{option}" in section "{section}" is not formatted as a list'
            )

    def getlistint(self, section, option, **kwargs):
        try:
            output = [int(x) for x in self.getlist(section, option, **kwargs)]
            return output
        except Exception as e:
            print(
                f'Some element in config option "{option}" in section "{section}" cannot be coerced into a integer'
            )
            raise e

    def getlistfloat(self, section, option, **kwargs):
        try:
            output = [float(x) for x in self.getlist(section, option, **kwargs)]
            return output
        except Exception as e:
            print(
                f'Some element in config option "{option}" in section "{section}" cannot be coerced into a float'
            )
            raise e


def is_saturated(
    channel: ndarray, threshold: float = 0.0, detector_bit_depth: Optional[int] = None
) -> bool:
    """
    Checks if the channel is saturated.
    A warning if it suspects that the detector bit depth does not match the datatype.
    thresh: float
        Threshold for the ratio of saturated pixels to total pixels
    detector_bit_depth: int
        Bit depth of the detector. Sometimes, detectors bit depth are not matching the datatype of the measureemnts.
        Here it can be specified the bit depth of the detector if known. The function is going to raise
        If None, it will be inferred from the channel dtype.
    """
    if detector_bit_depth is None:
        if np.issubdtype(channel.dtype, np.integer):
            max_limit = np.iinfo(channel.dtype).max
        elif np.issubdtype(channel.dtype, np.floating):
            max_limit = np.finfo(channel.dtype).max
        else:
            raise ValueError("The channel provided is not a valid numpy dtype.")

        for bit_depth in DETECTOR_BIT_DEPTHS:
            try:
                saturated = is_saturated(channel, threshold=threshold, detector_bit_depth=bit_depth)
            except ValueError:
                continue
        if saturated:
            warnings.warn(
                "The channel might be saturated. The channel saturation matches a common detector bit depth"
            )

    else:
        max_limit = pow(2, detector_bit_depth) - 1
        if np.any(channel > max_limit):
            raise ValueError(
                "The channel provided has values larger than the bit depth of the detector."
            )

    saturation_matrix = channel == max_limit
    saturation_ratio = np.count_nonzero(saturation_matrix) / channel.size

    return saturation_ratio > threshold



from datetime import datetime
from itertools import product
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from numpy import float64, int64, ndarray
from pandas import DataFrame
from scipy.interpolate import griddata
from scipy.optimize import curve_fit
from scipy.signal import find_peaks
from skimage.transform import hough_line  # hough_line_peaks, probabilistic_hough_line

import microscopemetrics.data_schema.samples.argolight_schema as schema
from microscopemetrics.analysis.tools import (
    compute_distances_matrix,
    compute_spots_properties,
    segment_image,
)
from microscopemetrics.data_schema import core_schema
from microscopemetrics.samples import AnalysisMixin, logger
from microscopemetrics.utilities.utilities import airy_fun, is_saturated, multi_airy_fun


class ArgolightBAnalysis(schema.ArgolightBDataset, AnalysisMixin):
    """This class handles the analysis of the Argolight sample pattern B"""

    def run(self) -> bool:
        self.validate_requirements()

        # Check image shape
        logger.info("Checking image shape...")
        image = self.input.argolight_b_image.data
        if len(image.shape) != 5:
            logger.error("Image must be 5D")
            return False

        # Check image saturation
        logger.info("Checking image saturation...")
        saturated_channels = []
        for c in range(image.shape[-1]):
            if is_saturated(
                channel=image[:, :, :, :, c],
                threshold=self.input.saturation_threshold,
                detector_bit_depth=self.input.bit_depth,
            ):
                logger.error(f"Channel {c} is saturated")
                saturated_channels.append(c)
        if len(saturated_channels):
            logger.error(f"Channels {saturated_channels} are saturated")
            return False

        # Calculating the distance between spots in pixels with a security margin
        min_distance = round(self.input.spots_distance * 0.3)

        # Calculating the maximum tolerated distance in microns for the same spot in a different channels
        max_distance = self.input.spots_distance * 0.4

        labels = segment_image(
            image=image,
            min_distance=min_distance,
            sigma=(self.input.sigma_z, self.input.sigma_y, self.input.sigma_x),
            method="local_max",
            low_corr_factors=self.input.lower_threshold_correction_factors,
            high_corr_factors=self.input.upper_threshold_correction_factors,
        )

        self.output.spots_labels_image = schema.ImageAsNumpy(  # TODO: this should be a mask
            data=labels,
            name=f"{self.input.argolight_b_image.name}_spots_labels",
            description=f"Spots labels of {self.input.argolight_b_image.image_url}",
            image_url=self.input.argolight_b_image.image_url,
            source_image_url=self.input.argolight_b_image.image_url,
        )

        spots_properties, spots_positions = compute_spots_properties(
            image=image,
            labels=labels,
            remove_center_cross=self.input.remove_center_cross,
        )

        distances_df = compute_distances_matrix(
            positions=spots_positions,
            max_distance=max_distance,
        )

        properties_kv = []
        properties_ls = []
        distances_kv = []
        spots_centroids = []

        for ch, ch_spot_props in enumerate(spots_properties):
            ch_df = DataFrame()
            ch_properties_kv = {}
            ch_df["channel"] = [ch for _ in ch_spot_props]
            ch_df["mask_labels"] = [p["label"] for p in ch_spot_props]
            ch_df["volume"] = [p["area"] for p in ch_spot_props]
            ch_df["roi_volume_units"] = "VOXEL"
            ch_df["max_intensity"] = [p["max_intensity"] for p in ch_spot_props]
            ch_df["min_intensity"] = [p["min_intensity"] for p in ch_spot_props]
            ch_df["mean_intensity"] = [p["mean_intensity"] for p in ch_spot_props]
            ch_df["integrated_intensity"] = [p["integrated_intensity"] for p in ch_spot_props]
            ch_df["z_weighted_centroid"] = [p["weighted_centroid"][0] for p in ch_spot_props]
            ch_df["y_weighted_centroid"] = [p["weighted_centroid"][1] for p in ch_spot_props]
            ch_df["x_weighted_centroid"] = [p["weighted_centroid"][2] for p in ch_spot_props]
            ch_df["roi_weighted_centroid_units"] = "PIXEL"

            # Key metrics for spots intensities
            ch_properties_kv["channel"] = ch
            ch_properties_kv["nr_of_spots"] = len(ch_df)
            ch_properties_kv["max_intensity"] = ch_df["integrated_intensity"].max().item()
            ch_properties_kv["max_intensity_roi"] = ch_df["integrated_intensity"].argmax().item()
            ch_properties_kv["min_intensity"] = ch_df["integrated_intensity"].min().item()
            ch_properties_kv["min_intensity_roi"] = ch_df["integrated_intensity"].argmin().item()
            ch_properties_kv["mean_intensity"] = ch_df["integrated_intensity"].mean().item()
            ch_properties_kv["median_intensity"] = ch_df["integrated_intensity"].median().item()
            ch_properties_kv["std_mean_intensity"] = ch_df["integrated_intensity"].std().item()
            ch_properties_kv["mad_mean_intensity"] = (
                (ch_df["integrated_intensity"] - ch_df["integrated_intensity"].mean()).abs().mean()
            )
            ch_properties_kv["min_max_intensity_ratio"] = (
                ch_properties_kv["min_intensity"] / ch_properties_kv["max_intensity"]
            )

            properties_ls.append(ch_df)
            properties_kv.append(ch_properties_kv)

            channel_shapes = [
                core_schema.Point(
                    x=p["weighted_centroid"][2].item(),
                    y=p["weighted_centroid"][1].item(),
                    z=p["weighted_centroid"][0].item(),
                    c=ch,
                    label=f'{p["label"]}',
                    # TODO: put some color
                )
                for p in ch_spot_props
            ]

            spots_centroids.append(
                schema.ROI(
                    label=f"Centroids_ch{ch:03d}",
                    image=self.input.argolight_b_image,
                    shapes=channel_shapes,
                )
            )
        properties_kv = {k: [i[k] for i in properties_kv] for k in properties_kv[0]}
        properties_df = pd.concat(properties_ls)

        for a, b in product(distances_df.channel_a.unique(), distances_df.channel_b.unique()):
            temp_df = distances_df[(distances_df.channel_a == a) & (distances_df.channel_b == b)]
            a = int(a)
            b = int(b)

            pr_distances_kv = {}
            pr_distances_kv["channel_A"] = a
            pr_distances_kv["channel_B"] = b
            pr_distances_kv["mean_3d_dist"] = temp_df.dist_3d.mean().item()
            pr_distances_kv["median_3d_dist"] = temp_df.dist_3d.median().item()
            pr_distances_kv["std_3d_dist"] = temp_df.dist_3d.std().item()
            pr_distances_kv["mad_3d_dist"] = (
                (temp_df.dist_3d - temp_df.dist_3d.mean()).abs().mean().item()
            )
            pr_distances_kv["mean_z_dist"] = temp_df.z_dist.mean().item()
            pr_distances_kv["median_z_dist"] = temp_df.z_dist.median().item()
            pr_distances_kv["std_z_dist"] = temp_df.z_dist.std().item()
            pr_distances_kv["mad_z_dist"] = (
                (temp_df.z_dist - temp_df.z_dist.mean()).abs().mean().item()
            )

            distances_kv.append(pr_distances_kv)

        distances_kv = {k: [i[k] for i in distances_kv] for k in distances_kv[0]}

        self.output.intensity_measurements = schema.ArgolightBIntensityKeyValues(**properties_kv)

        self.output.distance_measurements = schema.ArgolightBDistanceKeyValues(**distances_kv)

        self.output.spots_properties = schema.TableAsDict(
            columns=[
                core_schema.Column(name=k, values=v)
                for k, v in properties_df.to_dict(orient="list").items()
            ]
        )

        self.output.spots_distances = schema.TableAsDict(
            columns=[
                core_schema.Column(name=k, values=v)
                for k, v in distances_df.to_dict(orient="list").items()
            ]
        )

        self.output.spots_centroids = spots_centroids

        self.processing_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.processed = True

        return True


class ArgolightEAnalysis(schema.ArgolightEDataset, AnalysisMixin):
    """This class handles the analysis of the Argolight sample pattern E with lines along the X or Y axis"""

    def run(self) -> bool:
        self.validate_requirements()

        # Check image shape
        pass  # TODO: implement

        # Check image saturation
        pass  # TODO: implement

        # Check for axis value
        pass  # TODO: implement

        image = self.input.argolight_e_image.data
        image = np.squeeze(image[0])  # if there is a time dimension, take the first one
        axis = self.input.axis
        measured_band = self.input.measured_band

        (
            profiles,
            z_slices,
            peak_positions,
            peak_heights,
            peak_prominences,
            resolution_values,
            resolution_indexes,
        ) = _compute_resolution(
            image=image,
            axis=axis,
            measured_band=measured_band,
            prominence=self.input.prominence_threshold,
            do_angle_refinement=False,  # TODO: implement angle refinement
        )
        key_values = {
            "channel": [c for c in range(image.shape[-1])],
            "rayleigh_resolution": resolution_values,
            "peak_position_A": [
                peak_positions[ch][ind].item() for ch, ind in enumerate(resolution_indexes)
            ],
            "peak_position_B": [
                peak_positions[ch][ind + 1].item() for ch, ind in enumerate(resolution_indexes)
            ],
            "peak_height_A": [
                peak_heights[ch][ind].item() for ch, ind in enumerate(resolution_indexes)
            ],
            "peak_height_B": [
                peak_heights[ch][ind + 1].item() for ch, ind in enumerate(resolution_indexes)
            ],
            "peak_prominence_A": [
                peak_prominences[ch][ind].item() for ch, ind in enumerate(resolution_indexes)
            ],
            "peak_prominence_B": [
                peak_prominences[ch][ind + 1].item() for ch, ind in enumerate(resolution_indexes)
            ],
            "focus_slice": z_slices,
        }

        out_tables = []
        rois = []

        # Populate tables and rois
        for ch, profile in enumerate(profiles):
            out_tables.append(_profile_to_columns(profile, ch))
            shapes = []
            pos_a, pos_b = (
                key_values["peak_position_A"][ch],
                key_values["peak_position_B"][ch],
            )
            for peak in (pos_a, pos_b):
                # Measurements are taken at center of pixel so we add .5 pixel to peak positions
                if axis == 1:  # Y resolution -> horizontal rois
                    axis_len = image.shape[-1]
                    x1_pos = (axis_len / 2) - (axis_len * measured_band / 2)
                    y1_pos = peak + 0.5
                    x2_pos = (axis_len / 2) + (axis_len * measured_band / 2)
                    y2_pos = peak + 0.5
                elif axis == 2:  # X resolution -> vertical rois
                    axis_len = image.shape[-2]
                    y1_pos = (axis_len / 2) - (axis_len * measured_band / 2)
                    x1_pos = peak + 0.5
                    y2_pos = (axis_len / 2) + (axis_len * measured_band / 2)
                    x2_pos = peak + 0.5

                shapes.append(
                    core_schema.Line(
                        label=f"Ch_{ch}_resolution_{resolution_values[ch]:.2f}",
                        x1=x1_pos,
                        y1=y1_pos,
                        x2=x2_pos,
                        y2=y2_pos,
                        z=z_slices[ch],
                        c=ch,
                    )
                )
            rois.append(
                core_schema.ROI(
                    label=f"ch_{ch:03d}_peaks",
                    shapes=shapes,
                    image=self.input.argolight_e_image.image_url,
                )
            )
        self.output.peaks_rois = rois

        self.output.key_measurements = schema.ArgolightEKeyValues(**key_values)

        self.output.intensity_profiles = [schema.TableAsDict(columns=t) for t in out_tables]

        self.processing_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.processed = True

        return True


def _profile_to_columns(profile: ndarray, channel: int) -> list[dict[str, dict[str, list[float]]]]:
    table = [{f"raw_profile_ch{channel:02d}": {"values": [v.item() for v in profile[0, :]]}}]

    for p in range(1, profile.shape[0]):
        table.append(
            {
                f"fitted_profile_ch{channel:03d}_peak{p:03d}": {
                    "values": [v.item() for v in profile[p, :]]
                }
            }
        )

    return table


def _fit(
    profile: ndarray,
    peaks_guess: List[int64],
    amp: int = 4,
    exp: int = 2,
    lower_amp: int = 3,
    upper_amp: int = 5,
    center_tolerance: int = 1,
) -> Tuple[ndarray, ndarray, ndarray]:
    guess = []
    lower_bounds = []
    upper_bounds = []
    for p in peaks_guess:
        guess.append(p)  # peak center
        guess.append(amp)  # peak amplitude
        lower_bounds.append(p - center_tolerance)
        lower_bounds.append(lower_amp)
        upper_bounds.append(p + center_tolerance)
        upper_bounds.append(upper_amp)

    x = np.linspace(0, profile.shape[0], profile.shape[0], endpoint=False)

    popt, pcov = curve_fit(
        f=multi_airy_fun,
        xdata=x,
        ydata=profile,
        p0=guess,
        bounds=(lower_bounds, upper_bounds),
    )

    opt_peaks = popt[::2]
    # opt_amps = [a / 4 for a in popt[1::2]]  # We normalize back the amplitudes to the unity
    opt_amps = popt[1::2]

    fitted_profiles = np.zeros((len(peaks_guess), profile.shape[0]))
    for i, (c, a) in enumerate(zip(opt_peaks, opt_amps)):
        fitted_profiles[i, :] = airy_fun(x, c, a)

    return opt_peaks, opt_amps, fitted_profiles


def _compute_channel_resolution(
    channel: ndarray,
    axis: int,
    prominence_threshold: float,
    measured_band: float,
    do_fitting: bool = True,
    do_angle_refinement: bool = False,
) -> Tuple[ndarray, int64, ndarray, ndarray, ndarray, float64, int]:
    """Computes the resolution on a pattern of lines with increasing separation"""
    # find the most contrasted z-slice
    z_stdev = np.std(channel, axis=(1, 2))
    z_focus = np.argmax(z_stdev)
    focus_slice = channel[z_focus]

    # TODO: verify angle and correct
    if do_angle_refinement:
        # Set a precision of 0.1 degree.
        tested_angles = np.linspace(-np.pi / 2, np.pi / 2, 1800)
        h, theta, d = hough_line(focus_slice, theta=tested_angles)

    # Cut a band of that found peak
    # Best we can do now is just to cut a band in the center
    # We create a profiles along which we average signal
    axis_len = focus_slice.shape[-axis]
    weight_profile = np.zeros(axis_len)
    # Calculates a band of relative width 'image_fraction' to integrate the profile
    weight_profile[
        int((axis_len / 2) - (axis_len * measured_band / 2)) : int(
            (axis_len / 2) + (axis_len * measured_band / 2)
        )
    ] = 1
    profile = np.average(focus_slice, axis=-axis, weights=weight_profile)

    normalized_profile = (profile - np.min(profile)) / np.ptp(profile)

    peak_positions, properties = find_peaks(
        normalized_profile,
        height=0.3,
        distance=2,
        prominence=prominence_threshold / 4,
    )

    # From the properties we are interested in the amplitude
    # peak_heights = [h for h in properties['peak_heights']]
    ray_filtered_peak_pos = []
    ray_filtered_peak_heights = []
    ray_filtered_peak_prominences = []

    for peak, height, prom in zip(
        peak_positions, properties["peak_heights"], properties["prominences"]
    ):
        if (
            prom / height
        ) > prominence_threshold:  # This is calculating the prominence in relation to the local intensity
            ray_filtered_peak_pos.append(peak)
            ray_filtered_peak_heights.append(height)
            ray_filtered_peak_prominences.append(prom)

    peak_positions = ray_filtered_peak_pos
    peak_heights = ray_filtered_peak_heights
    peak_prominences = ray_filtered_peak_prominences

    if do_fitting:
        peak_positions, peak_heights, fitted_profiles = _fit(normalized_profile, peak_positions)
        normalized_profile = np.append(
            np.expand_dims(normalized_profile, 0), fitted_profiles, axis=0
        )

    # Find the closest peaks to return it as a measure of resolution
    peaks_distances = [abs(a - b) for a, b in zip(peak_positions[0:-2], peak_positions[1:-1])]
    if len(peaks_distances) == 0:
        res = None
        res_index = None
    else:
        res = min(peaks_distances)
        res_index = peaks_distances.index(res)

    return (
        normalized_profile,
        z_focus,
        peak_positions,
        peak_heights,
        peak_prominences,
        res,
        res_index,
    )


def _compute_resolution(
    image: ndarray,
    axis: int,
    measured_band: float,
    prominence: float,
    do_angle_refinement: bool = False,
) -> Tuple[
    List[ndarray],
    List[int],
    List[ndarray],
    List[ndarray],
    List[ndarray],
    List[float],
    List[int],
]:
    profiles = []
    z_planes = []
    peaks_positions = []
    peaks_heights = []
    peaks_prominences = []
    resolution_values = []
    resolution_indexes = []

    for c in range(image.shape[-1]):
        (
            prof,
            zp,
            pk_pos,
            pk_heights,
            pk_prominences,
            res,
            res_index,
        ) = _compute_channel_resolution(
            channel=image[..., c],
            axis=axis,
            prominence_threshold=prominence,
            measured_band=measured_band,
            do_angle_refinement=do_angle_refinement,
        )
        profiles.append(prof)
        z_planes.append(zp.item())
        peaks_positions.append(pk_pos)
        peaks_heights.append(pk_heights)
        peaks_prominences.append(pk_prominences)
        resolution_values.append(res.item())
        resolution_indexes.append(res_index)

    return (
        profiles,
        z_planes,
        peaks_positions,
        peaks_heights,
        peaks_prominences,
        resolution_values,
        resolution_indexes,
    )



import matplotlib.pyplot as plt
import numpy as np

from microscopemetrics.plots import Reporter


class ArgolightReporter(Reporter):
    """Reporter subclass to produce Argolight sample figures"""

    def __init__(self):
        image_report_to_func = {
            "spots": self.full_report_spots,
            "vertical_resolution": self.full_report_vertical_resolution,
            "horizontal_resolution": self.full_report_horizontal_resolution,
        }

        super().__init__(image_report_to_func=image_report_to_func)

    def produce_image_report(self, image):
        pass

    def full_report_spots(self, image):
        pass

    def full_report_vertical_resolution(self, image):
        pass

    def full_report_horizontal_resolution(self, image):
        pass

    def plot_homogeneity_map(self, image):
        nr_channels = image.getSizeC()
        x_dim = image.getSizeX()
        y_dim = image.getSizeY()

        tables = self.get_tables(
            image, namespace_start="metrics", name_filter="properties"
        )
        if len(tables) != 1:
            raise Exception(
                "There are none or more than one properties tables. Verify data integrity."
            )
        table = tables[0]

        row_count = table.getNumberOfRows()
        col_names = [c.name for c in table.getHeaders()]
        wanted_columns = [
            "channel",
            "max_intensity",
            "mean_intensity",
            "integrated_intensity",
            "x_weighted_centroid",
            "y_weighted_centroid",
        ]

        fig, axes = plt.subplots(
            ncols=nr_channels, nrows=3, squeeze=False, figsize=(3 * nr_channels, 9)
        )

        for ch in range(nr_channels):
            data = table.slice(
                [col_names.index(w_col) for w_col in wanted_columns],
                table.getWhereList(
                    condition=f"channel=={ch}",
                    variables={},
                    start=0,
                    stop=row_count,
                    step=0,
                ),
            )
            max_intensity = np.array(
                [
                    val
                    for col in data.columns
                    for val in col.values
                    if col.name == "max_intensity"
                ]
            )
            integrated_intensity = np.array(
                [
                    val
                    for col in data.columns
                    for val in col.values
                    if col.name == "integrated_intensity"
                ]
            )
            x_positions = np.array(
                [
                    val
                    for col in data.columns
                    for val in col.values
                    if col.name == "x_weighted_centroid"
                ]
            )
            y_positions = np.array(
                [
                    val
                    for col in data.columns
                    for val in col.values
                    if col.name == "y_weighted_centroid"
                ]
            )
            grid_x, grid_y = np.mgrid[0:x_dim, 0:y_dim]
            image_intensities = get_intensities(image, c_range=ch, t_range=0).max(0)

            try:
                interpolated_max_int = griddata(
                    np.stack((x_positions, y_positions), axis=1),
                    max_intensity,
                    (grid_x, grid_y),
                    method="linear",
                )
                interpolated_intgr_int = griddata(
                    np.stack((x_positions, y_positions), axis=1),
                    integrated_intensity,
                    (grid_x, grid_y),
                    method="linear",
                )
            except Exception as e:
                # TODO: Log a warning
                interpolated_max_int = np.zeros((256, 256))

            ax = axes.ravel()
            ax[ch] = plt.subplot(3, 4, ch + 1)

            ax[ch].imshow(np.squeeze(image_intensities), cmap="gray")
            ax[ch].set_title("MIP_" + str(ch))

            ax[ch + nr_channels].imshow(
                np.flipud(interpolated_intgr_int),
                extent=(0, x_dim, y_dim, 0),
                origin="lower",
                cmap=cm.hot,
                vmin=np.amin(integrated_intensity),
                vmax=np.amax(integrated_intensity),
            )
            ax[ch + nr_channels].plot(x_positions, y_positions, "k.", ms=2)
            ax[ch + nr_channels].set_title("Integrated_int_" + str(ch))

            ax[ch + 2 * nr_channels].imshow(
                np.flipud(interpolated_max_int),
                extent=(0, x_dim, y_dim, 0),
                origin="lower",
                cmap=cm.hot,
                vmin=np.amin(image_intensities),
                vmax=np.amax(image_intensities),
            )
            ax[ch + 2 * nr_channels].plot(x_positions, y_positions, "k.", ms=2)
            ax[ch + 2 * nr_channels].set_title("Max_int_" + str(ch))

        plt.show()

    def plot_distances_map(self, image):
        nr_channels = image.getSizeC()
        x_dim = image.getSizeX()
        y_dim = image.getSizeY()

        tables = get_tables(image, namespace_start="metrics", name_filter="distances")
        if len(tables) != 1:
            raise Exception(
                "There are none or more than one distances tables. Verify data integrity."
            )
        table = tables[0]
        row_count = table.getNumberOfRows()
        col_names = [c.name for c in table.getHeaders()]

        # We need the positions too
        pos_tables = get_tables(
            image, namespace_start="metrics", name_filter="properties"
        )
        if len(tables) != 1:
            raise Exception(
                "There are none or more than one positions tables. Verify data integrity."
            )
        pos_table = pos_tables[0]
        pos_row_count = pos_table.getNumberOfRows()
        pos_col_names = [c.name for c in pos_table.getHeaders()]

        fig, axes = plt.subplots(
            ncols=nr_channels - 1,
            nrows=nr_channels,
            squeeze=False,
            figsize=((nr_channels - 1) * 3, nr_channels * 3),
        )

        ax_index = 0
        for ch_A in range(nr_channels):
            pos_data = pos_table.slice(
                [
                    pos_col_names.index(w_col)
                    for w_col in [
                        "channel",
                        "mask_labels",
                        "x_weighted_centroid",
                        "y_weighted_centroid",
                    ]
                ],
                pos_table.getWhereList(
                    condition=f"channel=={ch_A}",
                    variables={},
                    start=0,
                    stop=pos_row_count,
                    step=0,
                ),
            )

            mask_labels = np.array(
                [
                    val
                    for col in pos_data.columns
                    for val in col.values
                    if col.name == "mask_labels"
                ]
            )
            x_positions = np.array(
                [
                    val
                    for col in pos_data.columns
                    for val in col.values
                    if col.name == "x_weighted_centroid"
                ]
            )
            y_positions = np.array(
                [
                    val
                    for col in pos_data.columns
                    for val in col.values
                    if col.name == "y_weighted_centroid"
                ]
            )
            positions_map = np.stack((x_positions, y_positions), axis=1)

            for ch_B in [i for i in range(nr_channels) if i != ch_A]:
                data = table.slice(
                    list(range(len(col_names))),
                    table.getWhereList(
                        condition=f"(channel_A=={ch_A})&(channel_B=={ch_B})",
                        variables={},
                        start=0,
                        stop=row_count,
                        step=0,
                    ),
                )
                labels_map = np.array(
                    [
                        val
                        for col in data.columns
                        for val in col.values
                        if col.name == "ch_A_roi_labels"
                    ]
                )
                labels_map += 1  # Mask labels are augmented by one as 0 is background
                distances_map_3d = np.array(
                    [
                        val
                        for col in data.columns
                        for val in col.values
                        if col.name == "distance_3d"
                    ]
                )
                distances_map_x = np.array(
                    [
                        val
                        for col in data.columns
                        for val in col.values
                        if col.name == "distance_x"
                    ]
                )
                distances_map_y = np.array(
                    [
                        val
                        for col in data.columns
                        for val in col.values
                        if col.name == "distance_y"
                    ]
                )
                distances_map_z = np.array(
                    [
                        val
                        for col in data.columns
                        for val in col.values
                        if col.name == "distance_z"
                    ]
                )

                filtered_positions = positions_map[
                    np.intersect1d(
                        mask_labels, labels_map, assume_unique=True, return_indices=True
                    )[1],
                    :,
                ]

                grid_x, grid_y = np.mgrid[0:x_dim:1, 0:y_dim:1]
                interpolated = griddata(
                    filtered_positions,
                    distances_map_3d,
                    (grid_x, grid_y),
                    method="cubic",
                )

                ax = axes.ravel()
                ax[ax_index].imshow(
                    np.flipud(interpolated),
                    extent=(0, x_dim, y_dim, 0),
                    origin="lower",
                    cmap=cm.hot,
                    vmin=np.amin(distances_map_3d),
                    vmax=np.amax(distances_map_3d),
                )
                ax[ax_index].set_title(f"Distance Ch{ch_A}-Ch{ch_B}")

                ax_index += 1

        plt.show()



# Auto generated from argolight_schema.yaml by pythongen.py version: 0.9.0
# Generation date: 2023-08-23T16:23:38
# Schema: microscopemetrics_samples_argolight_schema
#
# id: https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml
# description:
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import re
from dataclasses import dataclass
from typing import Any, ClassVar, Dict, List, Optional, Union

from jsonasobj2 import JsonObj, as_dict
from linkml_runtime.linkml_model.meta import (
    EnumDefinition,
    PermissibleValue,
    PvFormulaOptions,
)
from linkml_runtime.linkml_model.types import Boolean, Date, Float, Integer, String
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.utils.dataclass_extensions_376 import (
    dataclasses_init_fn_with_kwargs,
)
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from linkml_runtime.utils.formatutils import camelcase, sfx, underscore
from linkml_runtime.utils.metamodelcore import (
    Bool,
    XSDDate,
    bnode,
    empty_dict,
    empty_list,
)
from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.yamlutils import (
    YAMLRoot,
    extended_float,
    extended_int,
    extended_str,
)
from rdflib import Namespace, URIRef

from ..core_schema import (
    ROI,
    ExperimenterOrcid,
    ImageAsNumpy,
    ImageAsNumpyImageUrl,
    KeyValues,
    MetricsDataset,
    MetricsInput,
    MetricsOutput,
    SampleType,
    TableAsDict,
)

metamodel_version = "1.7.0"
version = None

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
LINKML = CurieNamespace("linkml", "https://w3id.org/linkml/")
DEFAULT_ = CurieNamespace(
    "",
    "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/",
)


# Types

# Class references


@dataclass
class ArgolightBDataset(MetricsDataset):
    """
    An Argolight sample pattern B dataset
    """

    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightBDataset"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "ArgolightBDataset"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightBDataset"
    )

    input: Union[dict, "ArgolightBInput"] = None
    processed: Union[bool, Bool] = False
    output: Optional[Union[dict, "ArgolightBOutput"]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.input):
            self.MissingRequiredField("input")
        if not isinstance(self.input, ArgolightBInput):
            self.input = ArgolightBInput(**as_dict(self.input))

        if self.output is not None and not isinstance(self.output, ArgolightBOutput):
            self.output = ArgolightBOutput(**as_dict(self.output))

        super().__post_init__(**kwargs)


@dataclass
class ArgolightEDataset(MetricsDataset):
    """
    An Argolight sample pattern E dataset.
    It contains resolution data on the axis indicated:
    - axis 1 = Y resolution = lines along X axis
    - axis 2 = X resolution = lines along Y axis
    """

    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightEDataset"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "ArgolightEDataset"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightEDataset"
    )

    input: Union[dict, "ArgolightEInput"] = None
    processed: Union[bool, Bool] = False
    output: Optional[Union[dict, "ArgolightEOutput"]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.input):
            self.MissingRequiredField("input")
        if not isinstance(self.input, ArgolightEInput):
            self.input = ArgolightEInput(**as_dict(self.input))

        if self.output is not None and not isinstance(self.output, ArgolightEOutput):
            self.output = ArgolightEOutput(**as_dict(self.output))

        super().__post_init__(**kwargs)


@dataclass
class ArgolightBInput(MetricsInput):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightBInput"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "ArgolightBInput"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightBInput"
    )

    argolight_b_image: Union[dict, ImageAsNumpy] = None
    spots_distance: float = None
    saturation_threshold: float = 0.01
    sigma_z: float = 1.0
    sigma_y: float = 3.0
    sigma_x: float = 3.0
    bit_depth: Optional[int] = None
    lower_threshold_correction_factors: Optional[Union[float, List[float]]] = empty_list()
    upper_threshold_correction_factors: Optional[Union[float, List[float]]] = empty_list()
    remove_center_cross: Optional[Union[bool, Bool]] = False

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.argolight_b_image):
            self.MissingRequiredField("argolight_b_image")
        if not isinstance(self.argolight_b_image, ImageAsNumpy):
            self.argolight_b_image = ImageAsNumpy(**as_dict(self.argolight_b_image))

        if self._is_empty(self.saturation_threshold):
            self.MissingRequiredField("saturation_threshold")
        if not isinstance(self.saturation_threshold, float):
            self.saturation_threshold = float(self.saturation_threshold)

        if self._is_empty(self.spots_distance):
            self.MissingRequiredField("spots_distance")
        if not isinstance(self.spots_distance, float):
            self.spots_distance = float(self.spots_distance)

        if self._is_empty(self.sigma_z):
            self.MissingRequiredField("sigma_z")
        if not isinstance(self.sigma_z, float):
            self.sigma_z = float(self.sigma_z)

        if self._is_empty(self.sigma_y):
            self.MissingRequiredField("sigma_y")
        if not isinstance(self.sigma_y, float):
            self.sigma_y = float(self.sigma_y)

        if self._is_empty(self.sigma_x):
            self.MissingRequiredField("sigma_x")
        if not isinstance(self.sigma_x, float):
            self.sigma_x = float(self.sigma_x)

        if self.bit_depth is not None and not isinstance(self.bit_depth, int):
            self.bit_depth = int(self.bit_depth)

        if not isinstance(self.lower_threshold_correction_factors, list):
            self.lower_threshold_correction_factors = (
                [self.lower_threshold_correction_factors]
                if self.lower_threshold_correction_factors is not None
                else []
            )
        self.lower_threshold_correction_factors = [
            v if isinstance(v, float) else float(v) for v in self.lower_threshold_correction_factors
        ]

        if not isinstance(self.upper_threshold_correction_factors, list):
            self.upper_threshold_correction_factors = (
                [self.upper_threshold_correction_factors]
                if self.upper_threshold_correction_factors is not None
                else []
            )
        self.upper_threshold_correction_factors = [
            v if isinstance(v, float) else float(v) for v in self.upper_threshold_correction_factors
        ]

        if self.remove_center_cross is not None and not isinstance(self.remove_center_cross, Bool):
            self.remove_center_cross = Bool(self.remove_center_cross)

        super().__post_init__(**kwargs)


@dataclass
class ArgolightEInput(MetricsInput):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightEInput"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "ArgolightEInput"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightEInput"
    )

    argolight_e_image: Union[dict, ImageAsNumpy] = None
    axis: int = None
    saturation_threshold: float = 0.01
    measured_band: float = 0.4
    prominence_threshold: float = 0.264
    bit_depth: Optional[int] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.argolight_e_image):
            self.MissingRequiredField("argolight_e_image")
        if not isinstance(self.argolight_e_image, ImageAsNumpy):
            self.argolight_e_image = ImageAsNumpy(**as_dict(self.argolight_e_image))

        if self._is_empty(self.saturation_threshold):
            self.MissingRequiredField("saturation_threshold")
        if not isinstance(self.saturation_threshold, float):
            self.saturation_threshold = float(self.saturation_threshold)

        if self._is_empty(self.axis):
            self.MissingRequiredField("axis")
        if not isinstance(self.axis, int):
            self.axis = int(self.axis)

        if self._is_empty(self.measured_band):
            self.MissingRequiredField("measured_band")
        if not isinstance(self.measured_band, float):
            self.measured_band = float(self.measured_band)

        if self._is_empty(self.prominence_threshold):
            self.MissingRequiredField("prominence_threshold")
        if not isinstance(self.prominence_threshold, float):
            self.prominence_threshold = float(self.prominence_threshold)

        if self.bit_depth is not None and not isinstance(self.bit_depth, int):
            self.bit_depth = int(self.bit_depth)

        super().__post_init__(**kwargs)


@dataclass
class ArgolightBOutput(MetricsOutput):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightBOutput"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "ArgolightBOutput"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightBOutput"
    )

    spots_labels_image: Optional[Union[str, ImageAsNumpyImageUrl]] = None
    spots_centroids: Optional[Union[Union[dict, ROI], List[Union[dict, ROI]]]] = empty_list()
    intensity_measurements: Optional[Union[dict, "ArgolightBIntensityKeyValues"]] = None
    distance_measurements: Optional[Union[dict, "ArgolightBDistanceKeyValues"]] = None
    spots_properties: Optional[Union[dict, TableAsDict]] = None
    spots_distances: Optional[Union[dict, TableAsDict]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.spots_labels_image is not None and not isinstance(
            self.spots_labels_image, ImageAsNumpyImageUrl
        ):
            self.spots_labels_image = ImageAsNumpyImageUrl(self.spots_labels_image)

        if not isinstance(self.spots_centroids, list):
            self.spots_centroids = (
                [self.spots_centroids] if self.spots_centroids is not None else []
            )
        self.spots_centroids = [
            v if isinstance(v, ROI) else ROI(**as_dict(v)) for v in self.spots_centroids
        ]

        if self.intensity_measurements is not None and not isinstance(
            self.intensity_measurements, ArgolightBIntensityKeyValues
        ):
            self.intensity_measurements = ArgolightBIntensityKeyValues(
                **as_dict(self.intensity_measurements)
            )

        if self.distance_measurements is not None and not isinstance(
            self.distance_measurements, ArgolightBDistanceKeyValues
        ):
            self.distance_measurements = ArgolightBDistanceKeyValues(
                **as_dict(self.distance_measurements)
            )

        if self.spots_properties is not None and not isinstance(self.spots_properties, TableAsDict):
            self.spots_properties = TableAsDict(**as_dict(self.spots_properties))

        if self.spots_distances is not None and not isinstance(self.spots_distances, TableAsDict):
            self.spots_distances = TableAsDict(**as_dict(self.spots_distances))

        super().__post_init__(**kwargs)


@dataclass
class ArgolightEOutput(MetricsOutput):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightEOutput"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "ArgolightEOutput"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightEOutput"
    )

    peaks_rois: Optional[Union[Union[dict, ROI], List[Union[dict, ROI]]]] = empty_list()
    key_measurements: Optional[Union[dict, "ArgolightEKeyValues"]] = None
    intensity_profiles: Optional[Union[dict, TableAsDict]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.peaks_rois, list):
            self.peaks_rois = [self.peaks_rois] if self.peaks_rois is not None else []
        self.peaks_rois = [v if isinstance(v, ROI) else ROI(**as_dict(v)) for v in self.peaks_rois]

        if self.key_measurements is not None and not isinstance(
            self.key_measurements, ArgolightEKeyValues
        ):
            self.key_measurements = ArgolightEKeyValues(**as_dict(self.key_measurements))

        if self.intensity_profiles is not None and not isinstance(
            self.intensity_profiles, TableAsDict
        ):
            self.intensity_profiles = TableAsDict(**as_dict(self.intensity_profiles))

        super().__post_init__(**kwargs)


@dataclass
class ArgolightBIntensityKeyValues(KeyValues):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightBIntensityKeyValues"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "ArgolightBIntensityKeyValues"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightBIntensityKeyValues"
    )

    channel: Optional[Union[int, List[int]]] = empty_list()
    nr_of_spots: Optional[Union[int, List[int]]] = empty_list()
    max_intensity: Optional[Union[float, List[float]]] = empty_list()
    max_intensity_roi: Optional[Union[int, List[int]]] = empty_list()
    min_intensity: Optional[Union[float, List[float]]] = empty_list()
    min_intensity_roi: Optional[Union[int, List[int]]] = empty_list()
    mean_intensity: Optional[Union[float, List[float]]] = empty_list()
    median_intensity: Optional[Union[float, List[float]]] = empty_list()
    std_mean_intensity: Optional[Union[float, List[float]]] = empty_list()
    mad_mean_intensity: Optional[Union[float, List[float]]] = empty_list()
    min_max_intensity_ratio: Optional[Union[float, List[float]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.channel, list):
            self.channel = [self.channel] if self.channel is not None else []
        self.channel = [v if isinstance(v, int) else int(v) for v in self.channel]

        if not isinstance(self.nr_of_spots, list):
            self.nr_of_spots = [self.nr_of_spots] if self.nr_of_spots is not None else []
        self.nr_of_spots = [v if isinstance(v, int) else int(v) for v in self.nr_of_spots]

        if not isinstance(self.max_intensity, list):
            self.max_intensity = [self.max_intensity] if self.max_intensity is not None else []
        self.max_intensity = [v if isinstance(v, float) else float(v) for v in self.max_intensity]

        if not isinstance(self.max_intensity_roi, list):
            self.max_intensity_roi = (
                [self.max_intensity_roi] if self.max_intensity_roi is not None else []
            )
        self.max_intensity_roi = [
            v if isinstance(v, int) else int(v) for v in self.max_intensity_roi
        ]

        if not isinstance(self.min_intensity, list):
            self.min_intensity = [self.min_intensity] if self.min_intensity is not None else []
        self.min_intensity = [v if isinstance(v, float) else float(v) for v in self.min_intensity]

        if not isinstance(self.min_intensity_roi, list):
            self.min_intensity_roi = (
                [self.min_intensity_roi] if self.min_intensity_roi is not None else []
            )
        self.min_intensity_roi = [
            v if isinstance(v, int) else int(v) for v in self.min_intensity_roi
        ]

        if not isinstance(self.mean_intensity, list):
            self.mean_intensity = [self.mean_intensity] if self.mean_intensity is not None else []
        self.mean_intensity = [v if isinstance(v, float) else float(v) for v in self.mean_intensity]

        if not isinstance(self.median_intensity, list):
            self.median_intensity = (
                [self.median_intensity] if self.median_intensity is not None else []
            )
        self.median_intensity = [
            v if isinstance(v, float) else float(v) for v in self.median_intensity
        ]

        if not isinstance(self.std_mean_intensity, list):
            self.std_mean_intensity = (
                [self.std_mean_intensity] if self.std_mean_intensity is not None else []
            )
        self.std_mean_intensity = [
            v if isinstance(v, float) else float(v) for v in self.std_mean_intensity
        ]

        if not isinstance(self.mad_mean_intensity, list):
            self.mad_mean_intensity = (
                [self.mad_mean_intensity] if self.mad_mean_intensity is not None else []
            )
        self.mad_mean_intensity = [
            v if isinstance(v, float) else float(v) for v in self.mad_mean_intensity
        ]

        if not isinstance(self.min_max_intensity_ratio, list):
            self.min_max_intensity_ratio = (
                [self.min_max_intensity_ratio] if self.min_max_intensity_ratio is not None else []
            )
        self.min_max_intensity_ratio = [
            v if isinstance(v, float) else float(v) for v in self.min_max_intensity_ratio
        ]

        super().__post_init__(**kwargs)


@dataclass
class ArgolightBDistanceKeyValues(KeyValues):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightBDistanceKeyValues"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "ArgolightBDistanceKeyValues"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightBDistanceKeyValues"
    )

    channel_A: Optional[Union[int, List[int]]] = empty_list()
    channel_B: Optional[Union[int, List[int]]] = empty_list()
    mean_3d_dist: Optional[Union[float, List[float]]] = empty_list()
    median_3d_dist: Optional[Union[float, List[float]]] = empty_list()
    std_3d_dist: Optional[Union[float, List[float]]] = empty_list()
    mad_3d_dist: Optional[Union[float, List[float]]] = empty_list()
    mean_z_dist: Optional[Union[float, List[float]]] = empty_list()
    median_z_dist: Optional[Union[float, List[float]]] = empty_list()
    std_z_dist: Optional[Union[float, List[float]]] = empty_list()
    mad_z_dist: Optional[Union[float, List[float]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.channel_A, list):
            self.channel_A = [self.channel_A] if self.channel_A is not None else []
        self.channel_A = [v if isinstance(v, int) else int(v) for v in self.channel_A]

        if not isinstance(self.channel_B, list):
            self.channel_B = [self.channel_B] if self.channel_B is not None else []
        self.channel_B = [v if isinstance(v, int) else int(v) for v in self.channel_B]

        if not isinstance(self.mean_3d_dist, list):
            self.mean_3d_dist = [self.mean_3d_dist] if self.mean_3d_dist is not None else []
        self.mean_3d_dist = [v if isinstance(v, float) else float(v) for v in self.mean_3d_dist]

        if not isinstance(self.median_3d_dist, list):
            self.median_3d_dist = [self.median_3d_dist] if self.median_3d_dist is not None else []
        self.median_3d_dist = [v if isinstance(v, float) else float(v) for v in self.median_3d_dist]

        if not isinstance(self.std_3d_dist, list):
            self.std_3d_dist = [self.std_3d_dist] if self.std_3d_dist is not None else []
        self.std_3d_dist = [v if isinstance(v, float) else float(v) for v in self.std_3d_dist]

        if not isinstance(self.mad_3d_dist, list):
            self.mad_3d_dist = [self.mad_3d_dist] if self.mad_3d_dist is not None else []
        self.mad_3d_dist = [v if isinstance(v, float) else float(v) for v in self.mad_3d_dist]

        if not isinstance(self.mean_z_dist, list):
            self.mean_z_dist = [self.mean_z_dist] if self.mean_z_dist is not None else []
        self.mean_z_dist = [v if isinstance(v, float) else float(v) for v in self.mean_z_dist]

        if not isinstance(self.median_z_dist, list):
            self.median_z_dist = [self.median_z_dist] if self.median_z_dist is not None else []
        self.median_z_dist = [v if isinstance(v, float) else float(v) for v in self.median_z_dist]

        if not isinstance(self.std_z_dist, list):
            self.std_z_dist = [self.std_z_dist] if self.std_z_dist is not None else []
        self.std_z_dist = [v if isinstance(v, float) else float(v) for v in self.std_z_dist]

        if not isinstance(self.mad_z_dist, list):
            self.mad_z_dist = [self.mad_z_dist] if self.mad_z_dist is not None else []
        self.mad_z_dist = [v if isinstance(v, float) else float(v) for v in self.mad_z_dist]

        super().__post_init__(**kwargs)


@dataclass
class ArgolightEKeyValues(KeyValues):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightEKeyValues"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "ArgolightEKeyValues"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/argolight_schema.yaml/ArgolightEKeyValues"
    )

    channel: Optional[Union[int, List[int]]] = empty_list()
    focus_slice: Optional[Union[int, List[int]]] = empty_list()
    rayleigh_resolution: Optional[Union[float, List[float]]] = empty_list()
    peak_position_A: Optional[Union[float, List[float]]] = empty_list()
    peak_position_B: Optional[Union[float, List[float]]] = empty_list()
    peak_height_A: Optional[Union[float, List[float]]] = empty_list()
    peak_height_B: Optional[Union[float, List[float]]] = empty_list()
    peak_prominence_A: Optional[Union[float, List[float]]] = empty_list()
    peak_prominence_B: Optional[Union[float, List[float]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.channel, list):
            self.channel = [self.channel] if self.channel is not None else []
        self.channel = [v if isinstance(v, int) else int(v) for v in self.channel]

        if not isinstance(self.focus_slice, list):
            self.focus_slice = [self.focus_slice] if self.focus_slice is not None else []
        self.focus_slice = [v if isinstance(v, int) else int(v) for v in self.focus_slice]

        if not isinstance(self.rayleigh_resolution, list):
            self.rayleigh_resolution = (
                [self.rayleigh_resolution] if self.rayleigh_resolution is not None else []
            )
        self.rayleigh_resolution = [
            v if isinstance(v, float) else float(v) for v in self.rayleigh_resolution
        ]

        if not isinstance(self.peak_position_A, list):
            self.peak_position_A = (
                [self.peak_position_A] if self.peak_position_A is not None else []
            )
        self.peak_position_A = [
            v if isinstance(v, float) else float(v) for v in self.peak_position_A
        ]

        if not isinstance(self.peak_position_B, list):
            self.peak_position_B = (
                [self.peak_position_B] if self.peak_position_B is not None else []
            )
        self.peak_position_B = [
            v if isinstance(v, float) else float(v) for v in self.peak_position_B
        ]

        if not isinstance(self.peak_height_A, list):
            self.peak_height_A = [self.peak_height_A] if self.peak_height_A is not None else []
        self.peak_height_A = [v if isinstance(v, float) else float(v) for v in self.peak_height_A]

        if not isinstance(self.peak_height_B, list):
            self.peak_height_B = [self.peak_height_B] if self.peak_height_B is not None else []
        self.peak_height_B = [v if isinstance(v, float) else float(v) for v in self.peak_height_B]

        if not isinstance(self.peak_prominence_A, list):
            self.peak_prominence_A = (
                [self.peak_prominence_A] if self.peak_prominence_A is not None else []
            )
        self.peak_prominence_A = [
            v if isinstance(v, float) else float(v) for v in self.peak_prominence_A
        ]

        if not isinstance(self.peak_prominence_B, list):
            self.peak_prominence_B = (
                [self.peak_prominence_B] if self.peak_prominence_B is not None else []
            )
        self.peak_prominence_B = [
            v if isinstance(v, float) else float(v) for v in self.peak_prominence_B
        ]

        super().__post_init__(**kwargs)


# Enumerations


# Slots
class slots:
    pass


slots.channel = Slot(
    uri=DEFAULT_.channel,
    name="channel",
    curie=DEFAULT_.curie("channel"),
    model_uri=DEFAULT_.channel,
    domain=None,
    range=Optional[Union[int, List[int]]],
)

slots.argolight_b_image = Slot(
    uri=DEFAULT_.argolight_b_image,
    name="argolight_b_image",
    curie=DEFAULT_.curie("argolight_b_image"),
    model_uri=DEFAULT_.argolight_b_image,
    domain=None,
    range=Union[dict, ImageAsNumpy],
)

slots.spots_distance = Slot(
    uri=DEFAULT_.spots_distance,
    name="spots_distance",
    curie=DEFAULT_.curie("spots_distance"),
    model_uri=DEFAULT_.spots_distance,
    domain=None,
    range=float,
)

slots.sigma_z = Slot(
    uri=DEFAULT_.sigma_z,
    name="sigma_z",
    curie=DEFAULT_.curie("sigma_z"),
    model_uri=DEFAULT_.sigma_z,
    domain=None,
    range=float,
)

slots.sigma_y = Slot(
    uri=DEFAULT_.sigma_y,
    name="sigma_y",
    curie=DEFAULT_.curie("sigma_y"),
    model_uri=DEFAULT_.sigma_y,
    domain=None,
    range=float,
)

slots.sigma_x = Slot(
    uri=DEFAULT_.sigma_x,
    name="sigma_x",
    curie=DEFAULT_.curie("sigma_x"),
    model_uri=DEFAULT_.sigma_x,
    domain=None,
    range=float,
)

slots.lower_threshold_correction_factors = Slot(
    uri=DEFAULT_.lower_threshold_correction_factors,
    name="lower_threshold_correction_factors",
    curie=DEFAULT_.curie("lower_threshold_correction_factors"),
    model_uri=DEFAULT_.lower_threshold_correction_factors,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.upper_threshold_correction_factors = Slot(
    uri=DEFAULT_.upper_threshold_correction_factors,
    name="upper_threshold_correction_factors",
    curie=DEFAULT_.curie("upper_threshold_correction_factors"),
    model_uri=DEFAULT_.upper_threshold_correction_factors,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.remove_center_cross = Slot(
    uri=DEFAULT_.remove_center_cross,
    name="remove_center_cross",
    curie=DEFAULT_.curie("remove_center_cross"),
    model_uri=DEFAULT_.remove_center_cross,
    domain=None,
    range=Optional[Union[bool, Bool]],
)

slots.nr_of_spots = Slot(
    uri=DEFAULT_.nr_of_spots,
    name="nr_of_spots",
    curie=DEFAULT_.curie("nr_of_spots"),
    model_uri=DEFAULT_.nr_of_spots,
    domain=None,
    range=Optional[Union[int, List[int]]],
)

slots.max_intensity = Slot(
    uri=DEFAULT_.max_intensity,
    name="max_intensity",
    curie=DEFAULT_.curie("max_intensity"),
    model_uri=DEFAULT_.max_intensity,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.max_intensity_roi = Slot(
    uri=DEFAULT_.max_intensity_roi,
    name="max_intensity_roi",
    curie=DEFAULT_.curie("max_intensity_roi"),
    model_uri=DEFAULT_.max_intensity_roi,
    domain=None,
    range=Optional[Union[int, List[int]]],
)

slots.min_intensity = Slot(
    uri=DEFAULT_.min_intensity,
    name="min_intensity",
    curie=DEFAULT_.curie("min_intensity"),
    model_uri=DEFAULT_.min_intensity,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.min_intensity_roi = Slot(
    uri=DEFAULT_.min_intensity_roi,
    name="min_intensity_roi",
    curie=DEFAULT_.curie("min_intensity_roi"),
    model_uri=DEFAULT_.min_intensity_roi,
    domain=None,
    range=Optional[Union[int, List[int]]],
)

slots.mean_intensity = Slot(
    uri=DEFAULT_.mean_intensity,
    name="mean_intensity",
    curie=DEFAULT_.curie("mean_intensity"),
    model_uri=DEFAULT_.mean_intensity,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.median_intensity = Slot(
    uri=DEFAULT_.median_intensity,
    name="median_intensity",
    curie=DEFAULT_.curie("median_intensity"),
    model_uri=DEFAULT_.median_intensity,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.std_mean_intensity = Slot(
    uri=DEFAULT_.std_mean_intensity,
    name="std_mean_intensity",
    curie=DEFAULT_.curie("std_mean_intensity"),
    model_uri=DEFAULT_.std_mean_intensity,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.mad_mean_intensity = Slot(
    uri=DEFAULT_.mad_mean_intensity,
    name="mad_mean_intensity",
    curie=DEFAULT_.curie("mad_mean_intensity"),
    model_uri=DEFAULT_.mad_mean_intensity,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.min_max_intensity_ratio = Slot(
    uri=DEFAULT_.min_max_intensity_ratio,
    name="min_max_intensity_ratio",
    curie=DEFAULT_.curie("min_max_intensity_ratio"),
    model_uri=DEFAULT_.min_max_intensity_ratio,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.channel_A = Slot(
    uri=DEFAULT_.channel_A,
    name="channel_A",
    curie=DEFAULT_.curie("channel_A"),
    model_uri=DEFAULT_.channel_A,
    domain=None,
    range=Optional[Union[int, List[int]]],
)

slots.channel_B = Slot(
    uri=DEFAULT_.channel_B,
    name="channel_B",
    curie=DEFAULT_.curie("channel_B"),
    model_uri=DEFAULT_.channel_B,
    domain=None,
    range=Optional[Union[int, List[int]]],
)

slots.mean_3d_dist = Slot(
    uri=DEFAULT_.mean_3d_dist,
    name="mean_3d_dist",
    curie=DEFAULT_.curie("mean_3d_dist"),
    model_uri=DEFAULT_.mean_3d_dist,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.median_3d_dist = Slot(
    uri=DEFAULT_.median_3d_dist,
    name="median_3d_dist",
    curie=DEFAULT_.curie("median_3d_dist"),
    model_uri=DEFAULT_.median_3d_dist,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.std_3d_dist = Slot(
    uri=DEFAULT_.std_3d_dist,
    name="std_3d_dist",
    curie=DEFAULT_.curie("std_3d_dist"),
    model_uri=DEFAULT_.std_3d_dist,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.mad_3d_dist = Slot(
    uri=DEFAULT_.mad_3d_dist,
    name="mad_3d_dist",
    curie=DEFAULT_.curie("mad_3d_dist"),
    model_uri=DEFAULT_.mad_3d_dist,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.mean_z_dist = Slot(
    uri=DEFAULT_.mean_z_dist,
    name="mean_z_dist",
    curie=DEFAULT_.curie("mean_z_dist"),
    model_uri=DEFAULT_.mean_z_dist,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.median_z_dist = Slot(
    uri=DEFAULT_.median_z_dist,
    name="median_z_dist",
    curie=DEFAULT_.curie("median_z_dist"),
    model_uri=DEFAULT_.median_z_dist,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.std_z_dist = Slot(
    uri=DEFAULT_.std_z_dist,
    name="std_z_dist",
    curie=DEFAULT_.curie("std_z_dist"),
    model_uri=DEFAULT_.std_z_dist,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.mad_z_dist = Slot(
    uri=DEFAULT_.mad_z_dist,
    name="mad_z_dist",
    curie=DEFAULT_.curie("mad_z_dist"),
    model_uri=DEFAULT_.mad_z_dist,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.argolight_e_image = Slot(
    uri=DEFAULT_.argolight_e_image,
    name="argolight_e_image",
    curie=DEFAULT_.curie("argolight_e_image"),
    model_uri=DEFAULT_.argolight_e_image,
    domain=None,
    range=Union[dict, ImageAsNumpy],
)

slots.axis = Slot(
    uri=DEFAULT_.axis,
    name="axis",
    curie=DEFAULT_.curie("axis"),
    model_uri=DEFAULT_.axis,
    domain=None,
    range=int,
)

slots.measured_band = Slot(
    uri=DEFAULT_.measured_band,
    name="measured_band",
    curie=DEFAULT_.curie("measured_band"),
    model_uri=DEFAULT_.measured_band,
    domain=None,
    range=float,
)

slots.prominence_threshold = Slot(
    uri=DEFAULT_.prominence_threshold,
    name="prominence_threshold",
    curie=DEFAULT_.curie("prominence_threshold"),
    model_uri=DEFAULT_.prominence_threshold,
    domain=None,
    range=float,
)

slots.focus_slice = Slot(
    uri=DEFAULT_.focus_slice,
    name="focus_slice",
    curie=DEFAULT_.curie("focus_slice"),
    model_uri=DEFAULT_.focus_slice,
    domain=None,
    range=Optional[Union[int, List[int]]],
)

slots.rayleigh_resolution = Slot(
    uri=DEFAULT_.rayleigh_resolution,
    name="rayleigh_resolution",
    curie=DEFAULT_.curie("rayleigh_resolution"),
    model_uri=DEFAULT_.rayleigh_resolution,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.peak_position_A = Slot(
    uri=DEFAULT_.peak_position_A,
    name="peak_position_A",
    curie=DEFAULT_.curie("peak_position_A"),
    model_uri=DEFAULT_.peak_position_A,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.peak_position_B = Slot(
    uri=DEFAULT_.peak_position_B,
    name="peak_position_B",
    curie=DEFAULT_.curie("peak_position_B"),
    model_uri=DEFAULT_.peak_position_B,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.peak_height_A = Slot(
    uri=DEFAULT_.peak_height_A,
    name="peak_height_A",
    curie=DEFAULT_.curie("peak_height_A"),
    model_uri=DEFAULT_.peak_height_A,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.peak_height_B = Slot(
    uri=DEFAULT_.peak_height_B,
    name="peak_height_B",
    curie=DEFAULT_.curie("peak_height_B"),
    model_uri=DEFAULT_.peak_height_B,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.peak_prominence_A = Slot(
    uri=DEFAULT_.peak_prominence_A,
    name="peak_prominence_A",
    curie=DEFAULT_.curie("peak_prominence_A"),
    model_uri=DEFAULT_.peak_prominence_A,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.peak_prominence_B = Slot(
    uri=DEFAULT_.peak_prominence_B,
    name="peak_prominence_B",
    curie=DEFAULT_.curie("peak_prominence_B"),
    model_uri=DEFAULT_.peak_prominence_B,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.argolightBDataset__input = Slot(
    uri=DEFAULT_.input,
    name="argolightBDataset__input",
    curie=DEFAULT_.curie("input"),
    model_uri=DEFAULT_.argolightBDataset__input,
    domain=None,
    range=Union[dict, ArgolightBInput],
)

slots.argolightBDataset__output = Slot(
    uri=DEFAULT_.output,
    name="argolightBDataset__output",
    curie=DEFAULT_.curie("output"),
    model_uri=DEFAULT_.argolightBDataset__output,
    domain=None,
    range=Optional[Union[dict, ArgolightBOutput]],
)

slots.argolightBOutput__spots_labels_image = Slot(
    uri=DEFAULT_.spots_labels_image,
    name="argolightBOutput__spots_labels_image",
    curie=DEFAULT_.curie("spots_labels_image"),
    model_uri=DEFAULT_.argolightBOutput__spots_labels_image,
    domain=None,
    range=Optional[Union[str, ImageAsNumpyImageUrl]],
)

slots.argolightBOutput__spots_centroids = Slot(
    uri=DEFAULT_.spots_centroids,
    name="argolightBOutput__spots_centroids",
    curie=DEFAULT_.curie("spots_centroids"),
    model_uri=DEFAULT_.argolightBOutput__spots_centroids,
    domain=None,
    range=Optional[Union[Union[dict, ROI], List[Union[dict, ROI]]]],
)

slots.argolightBOutput__intensity_measurements = Slot(
    uri=DEFAULT_.intensity_measurements,
    name="argolightBOutput__intensity_measurements",
    curie=DEFAULT_.curie("intensity_measurements"),
    model_uri=DEFAULT_.argolightBOutput__intensity_measurements,
    domain=None,
    range=Optional[Union[dict, ArgolightBIntensityKeyValues]],
)

slots.argolightBOutput__distance_measurements = Slot(
    uri=DEFAULT_.distance_measurements,
    name="argolightBOutput__distance_measurements",
    curie=DEFAULT_.curie("distance_measurements"),
    model_uri=DEFAULT_.argolightBOutput__distance_measurements,
    domain=None,
    range=Optional[Union[dict, ArgolightBDistanceKeyValues]],
)

slots.argolightBOutput__spots_properties = Slot(
    uri=DEFAULT_.spots_properties,
    name="argolightBOutput__spots_properties",
    curie=DEFAULT_.curie("spots_properties"),
    model_uri=DEFAULT_.argolightBOutput__spots_properties,
    domain=None,
    range=Optional[Union[dict, TableAsDict]],
)

slots.argolightBOutput__spots_distances = Slot(
    uri=DEFAULT_.spots_distances,
    name="argolightBOutput__spots_distances",
    curie=DEFAULT_.curie("spots_distances"),
    model_uri=DEFAULT_.argolightBOutput__spots_distances,
    domain=None,
    range=Optional[Union[dict, TableAsDict]],
)

slots.argolightEDataset__input = Slot(
    uri=DEFAULT_.input,
    name="argolightEDataset__input",
    curie=DEFAULT_.curie("input"),
    model_uri=DEFAULT_.argolightEDataset__input,
    domain=None,
    range=Union[dict, ArgolightEInput],
)

slots.argolightEDataset__output = Slot(
    uri=DEFAULT_.output,
    name="argolightEDataset__output",
    curie=DEFAULT_.curie("output"),
    model_uri=DEFAULT_.argolightEDataset__output,
    domain=None,
    range=Optional[Union[dict, ArgolightEOutput]],
)

slots.argolightEOutput__peaks_rois = Slot(
    uri=DEFAULT_.peaks_rois,
    name="argolightEOutput__peaks_rois",
    curie=DEFAULT_.curie("peaks_rois"),
    model_uri=DEFAULT_.argolightEOutput__peaks_rois,
    domain=None,
    range=Optional[Union[Union[dict, ROI], List[Union[dict, ROI]]]],
)

slots.argolightEOutput__key_measurements = Slot(
    uri=DEFAULT_.key_measurements,
    name="argolightEOutput__key_measurements",
    curie=DEFAULT_.curie("key_measurements"),
    model_uri=DEFAULT_.argolightEOutput__key_measurements,
    domain=None,
    range=Optional[Union[dict, ArgolightEKeyValues]],
)

slots.argolightEOutput__intensity_profiles = Slot(
    uri=DEFAULT_.intensity_profiles,
    name="argolightEOutput__intensity_profiles",
    curie=DEFAULT_.curie("intensity_profiles"),
    model_uri=DEFAULT_.argolightEOutput__intensity_profiles,
    domain=None,
    range=Optional[Union[dict, TableAsDict]],
)



# Auto generated from field_illumination_schema.yaml by pythongen.py version: 0.9.0
# Generation date: 2023-08-23T16:23:29
# Schema: microscopemetrics_samples_field_illumination_schema
#
# id: https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml
# description:
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import re
from dataclasses import dataclass
from typing import Any, ClassVar, Dict, List, Optional, Union

from jsonasobj2 import JsonObj, as_dict
from linkml_runtime.linkml_model.meta import (
    EnumDefinition,
    PermissibleValue,
    PvFormulaOptions,
)
from linkml_runtime.linkml_model.types import Boolean, Date, Float, Integer, String
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.utils.dataclass_extensions_376 import (
    dataclasses_init_fn_with_kwargs,
)
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from linkml_runtime.utils.formatutils import camelcase, sfx, underscore
from linkml_runtime.utils.metamodelcore import (
    Bool,
    XSDDate,
    bnode,
    empty_dict,
    empty_list,
)
from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.yamlutils import (
    YAMLRoot,
    extended_float,
    extended_int,
    extended_str,
)
from rdflib import Namespace, URIRef

from ..core_schema import (
    ROI,
    ExperimenterOrcid,
    Image5DImageUrl,
    ImageAsNumpy,
    ImageAsNumpyImageUrl,
    KeyValues,
    MetricsDataset,
    MetricsInput,
    MetricsOutput,
    SampleType,
    TableAsDict,
)

metamodel_version = "1.7.0"
version = None

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
LINKML = CurieNamespace("linkml", "https://w3id.org/linkml/")
DEFAULT_ = CurieNamespace(
    "",
    "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/",
)


# Types

# Class references


@dataclass
class FieldIlluminationDataset(MetricsDataset):
    """
    A field illumination dataset
    """

    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/FieldIlluminationDataset"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "FieldIlluminationDataset"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/FieldIlluminationDataset"
    )

    input: Union[dict, "FieldIlluminationInput"] = None
    processed: Union[bool, Bool] = False
    output: Optional[Union[dict, "FieldIlluminationOutput"]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.input):
            self.MissingRequiredField("input")
        if not isinstance(self.input, FieldIlluminationInput):
            self.input = FieldIlluminationInput(**as_dict(self.input))

        if self.output is not None and not isinstance(self.output, FieldIlluminationOutput):
            self.output = FieldIlluminationOutput(**as_dict(self.output))

        super().__post_init__(**kwargs)


@dataclass
class FieldIlluminationInput(MetricsInput):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/FieldIlluminationInput"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "FieldIlluminationInput"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/FieldIlluminationInput"
    )

    field_illumination_image: Union[dict, ImageAsNumpy] = None
    saturation_threshold: float = 0.01
    center_threshold: float = 0.9
    corner_fraction: float = 0.1
    sigma: float = 2.0
    intensity_map_size: int = 64
    bit_depth: Optional[int] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.field_illumination_image):
            self.MissingRequiredField("field_illumination_image")
        if not isinstance(self.field_illumination_image, ImageAsNumpy):
            self.field_illumination_image = ImageAsNumpy(**as_dict(self.field_illumination_image))

        if self._is_empty(self.saturation_threshold):
            self.MissingRequiredField("saturation_threshold")
        if not isinstance(self.saturation_threshold, float):
            self.saturation_threshold = float(self.saturation_threshold)

        if self._is_empty(self.center_threshold):
            self.MissingRequiredField("center_threshold")
        if not isinstance(self.center_threshold, float):
            self.center_threshold = float(self.center_threshold)

        if self._is_empty(self.corner_fraction):
            self.MissingRequiredField("corner_fraction")
        if not isinstance(self.corner_fraction, float):
            self.corner_fraction = float(self.corner_fraction)

        if self._is_empty(self.sigma):
            self.MissingRequiredField("sigma")
        if not isinstance(self.sigma, float):
            self.sigma = float(self.sigma)

        if self._is_empty(self.intensity_map_size):
            self.MissingRequiredField("intensity_map_size")
        if not isinstance(self.intensity_map_size, int):
            self.intensity_map_size = int(self.intensity_map_size)

        if self.bit_depth is not None and not isinstance(self.bit_depth, int):
            self.bit_depth = int(self.bit_depth)

        super().__post_init__(**kwargs)


@dataclass
class FieldIlluminationOutput(MetricsOutput):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/FieldIlluminationOutput"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "FieldIlluminationOutput"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/FieldIlluminationOutput"
    )

    key_values: Optional[Union[dict, "FieldIlluminationKeyValues"]] = None
    intensity_profiles: Optional[Union[dict, TableAsDict]] = None
    intensity_map: Optional[Union[str, Image5DImageUrl]] = None
    profile_rois: Optional[Union[dict, ROI]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.key_values is not None and not isinstance(
            self.key_values, FieldIlluminationKeyValues
        ):
            self.key_values = FieldIlluminationKeyValues(**as_dict(self.key_values))

        if self.intensity_profiles is not None and not isinstance(
            self.intensity_profiles, TableAsDict
        ):
            self.intensity_profiles = TableAsDict(**as_dict(self.intensity_profiles))

        if self.intensity_map is not None and not isinstance(self.intensity_map, Image5DImageUrl):
            self.intensity_map = Image5DImageUrl(self.intensity_map)

        if self.profile_rois is not None and not isinstance(self.profile_rois, ROI):
            self.profile_rois = ROI(**as_dict(self.profile_rois))

        super().__post_init__(**kwargs)


@dataclass
class FieldIlluminationKeyValues(KeyValues):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/FieldIlluminationKeyValues"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "FieldIlluminationKeyValues"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/FieldIlluminationKeyValues"
    )

    channel: Optional[Union[int, List[int]]] = empty_list()
    nb_pixels: Optional[Union[int, List[int]]] = empty_list()
    center_of_mass_x: Optional[Union[float, List[float]]] = empty_list()
    center_of_mass_y: Optional[Union[float, List[float]]] = empty_list()
    max_intensity: Optional[Union[float, List[float]]] = empty_list()
    max_intensity_pos_x: Optional[Union[float, List[float]]] = empty_list()
    max_intensity_pos_y: Optional[Union[float, List[float]]] = empty_list()
    top_left_intensity_mean: Optional[Union[float, List[float]]] = empty_list()
    top_left_intensity_ratio: Optional[Union[float, List[float]]] = empty_list()
    top_center_intensity_mean: Optional[Union[float, List[float]]] = empty_list()
    top_center_intensity_ratio: Optional[Union[float, List[float]]] = empty_list()
    top_right_intensity_mean: Optional[Union[float, List[float]]] = empty_list()
    top_right_intensity_ratio: Optional[Union[float, List[float]]] = empty_list()
    middle_left_intensity_mean: Optional[Union[float, List[float]]] = empty_list()
    middle_left_intensity_ratio: Optional[Union[float, List[float]]] = empty_list()
    middle_center_intensity_mean: Optional[Union[float, List[float]]] = empty_list()
    middle_center_intensity_ratio: Optional[Union[float, List[float]]] = empty_list()
    middle_right_intensity_mean: Optional[Union[float, List[float]]] = empty_list()
    middle_right_intensity_ratio: Optional[Union[float, List[float]]] = empty_list()
    bottom_left_intensity_mean: Optional[Union[float, List[float]]] = empty_list()
    bottom_left_intensity_ratio: Optional[Union[float, List[float]]] = empty_list()
    bottom_center_intensity_mean: Optional[Union[float, List[float]]] = empty_list()
    bottom_center_intensity_ratio: Optional[Union[float, List[float]]] = empty_list()
    bottom_right_intensity_mean: Optional[Union[float, List[float]]] = empty_list()
    bottom_right_intensity_ratio: Optional[Union[float, List[float]]] = empty_list()
    decile_0: Optional[Union[float, List[float]]] = empty_list()
    decile_1: Optional[Union[float, List[float]]] = empty_list()
    decile_2: Optional[Union[float, List[float]]] = empty_list()
    decile_3: Optional[Union[float, List[float]]] = empty_list()
    decile_4: Optional[Union[float, List[float]]] = empty_list()
    decile_5: Optional[Union[float, List[float]]] = empty_list()
    decile_6: Optional[Union[float, List[float]]] = empty_list()
    decile_7: Optional[Union[float, List[float]]] = empty_list()
    decile_8: Optional[Union[float, List[float]]] = empty_list()
    decile_9: Optional[Union[float, List[float]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.channel, list):
            self.channel = [self.channel] if self.channel is not None else []
        self.channel = [v if isinstance(v, int) else int(v) for v in self.channel]

        if not isinstance(self.nb_pixels, list):
            self.nb_pixels = [self.nb_pixels] if self.nb_pixels is not None else []
        self.nb_pixels = [v if isinstance(v, int) else int(v) for v in self.nb_pixels]

        if not isinstance(self.center_of_mass_x, list):
            self.center_of_mass_x = (
                [self.center_of_mass_x] if self.center_of_mass_x is not None else []
            )
        self.center_of_mass_x = [
            v if isinstance(v, float) else float(v) for v in self.center_of_mass_x
        ]

        if not isinstance(self.center_of_mass_y, list):
            self.center_of_mass_y = (
                [self.center_of_mass_y] if self.center_of_mass_y is not None else []
            )
        self.center_of_mass_y = [
            v if isinstance(v, float) else float(v) for v in self.center_of_mass_y
        ]

        if not isinstance(self.max_intensity, list):
            self.max_intensity = [self.max_intensity] if self.max_intensity is not None else []
        self.max_intensity = [v if isinstance(v, float) else float(v) for v in self.max_intensity]

        if not isinstance(self.max_intensity_pos_x, list):
            self.max_intensity_pos_x = (
                [self.max_intensity_pos_x] if self.max_intensity_pos_x is not None else []
            )
        self.max_intensity_pos_x = [
            v if isinstance(v, float) else float(v) for v in self.max_intensity_pos_x
        ]

        if not isinstance(self.max_intensity_pos_y, list):
            self.max_intensity_pos_y = (
                [self.max_intensity_pos_y] if self.max_intensity_pos_y is not None else []
            )
        self.max_intensity_pos_y = [
            v if isinstance(v, float) else float(v) for v in self.max_intensity_pos_y
        ]

        if not isinstance(self.top_left_intensity_mean, list):
            self.top_left_intensity_mean = (
                [self.top_left_intensity_mean] if self.top_left_intensity_mean is not None else []
            )
        self.top_left_intensity_mean = [
            v if isinstance(v, float) else float(v) for v in self.top_left_intensity_mean
        ]

        if not isinstance(self.top_left_intensity_ratio, list):
            self.top_left_intensity_ratio = (
                [self.top_left_intensity_ratio] if self.top_left_intensity_ratio is not None else []
            )
        self.top_left_intensity_ratio = [
            v if isinstance(v, float) else float(v) for v in self.top_left_intensity_ratio
        ]

        if not isinstance(self.top_center_intensity_mean, list):
            self.top_center_intensity_mean = (
                [self.top_center_intensity_mean]
                if self.top_center_intensity_mean is not None
                else []
            )
        self.top_center_intensity_mean = [
            v if isinstance(v, float) else float(v) for v in self.top_center_intensity_mean
        ]

        if not isinstance(self.top_center_intensity_ratio, list):
            self.top_center_intensity_ratio = (
                [self.top_center_intensity_ratio]
                if self.top_center_intensity_ratio is not None
                else []
            )
        self.top_center_intensity_ratio = [
            v if isinstance(v, float) else float(v) for v in self.top_center_intensity_ratio
        ]

        if not isinstance(self.top_right_intensity_mean, list):
            self.top_right_intensity_mean = (
                [self.top_right_intensity_mean] if self.top_right_intensity_mean is not None else []
            )
        self.top_right_intensity_mean = [
            v if isinstance(v, float) else float(v) for v in self.top_right_intensity_mean
        ]

        if not isinstance(self.top_right_intensity_ratio, list):
            self.top_right_intensity_ratio = (
                [self.top_right_intensity_ratio]
                if self.top_right_intensity_ratio is not None
                else []
            )
        self.top_right_intensity_ratio = [
            v if isinstance(v, float) else float(v) for v in self.top_right_intensity_ratio
        ]

        if not isinstance(self.middle_left_intensity_mean, list):
            self.middle_left_intensity_mean = (
                [self.middle_left_intensity_mean]
                if self.middle_left_intensity_mean is not None
                else []
            )
        self.middle_left_intensity_mean = [
            v if isinstance(v, float) else float(v) for v in self.middle_left_intensity_mean
        ]

        if not isinstance(self.middle_left_intensity_ratio, list):
            self.middle_left_intensity_ratio = (
                [self.middle_left_intensity_ratio]
                if self.middle_left_intensity_ratio is not None
                else []
            )
        self.middle_left_intensity_ratio = [
            v if isinstance(v, float) else float(v) for v in self.middle_left_intensity_ratio
        ]

        if not isinstance(self.middle_center_intensity_mean, list):
            self.middle_center_intensity_mean = (
                [self.middle_center_intensity_mean]
                if self.middle_center_intensity_mean is not None
                else []
            )
        self.middle_center_intensity_mean = [
            v if isinstance(v, float) else float(v) for v in self.middle_center_intensity_mean
        ]

        if not isinstance(self.middle_center_intensity_ratio, list):
            self.middle_center_intensity_ratio = (
                [self.middle_center_intensity_ratio]
                if self.middle_center_intensity_ratio is not None
                else []
            )
        self.middle_center_intensity_ratio = [
            v if isinstance(v, float) else float(v) for v in self.middle_center_intensity_ratio
        ]

        if not isinstance(self.middle_right_intensity_mean, list):
            self.middle_right_intensity_mean = (
                [self.middle_right_intensity_mean]
                if self.middle_right_intensity_mean is not None
                else []
            )
        self.middle_right_intensity_mean = [
            v if isinstance(v, float) else float(v) for v in self.middle_right_intensity_mean
        ]

        if not isinstance(self.middle_right_intensity_ratio, list):
            self.middle_right_intensity_ratio = (
                [self.middle_right_intensity_ratio]
                if self.middle_right_intensity_ratio is not None
                else []
            )
        self.middle_right_intensity_ratio = [
            v if isinstance(v, float) else float(v) for v in self.middle_right_intensity_ratio
        ]

        if not isinstance(self.bottom_left_intensity_mean, list):
            self.bottom_left_intensity_mean = (
                [self.bottom_left_intensity_mean]
                if self.bottom_left_intensity_mean is not None
                else []
            )
        self.bottom_left_intensity_mean = [
            v if isinstance(v, float) else float(v) for v in self.bottom_left_intensity_mean
        ]

        if not isinstance(self.bottom_left_intensity_ratio, list):
            self.bottom_left_intensity_ratio = (
                [self.bottom_left_intensity_ratio]
                if self.bottom_left_intensity_ratio is not None
                else []
            )
        self.bottom_left_intensity_ratio = [
            v if isinstance(v, float) else float(v) for v in self.bottom_left_intensity_ratio
        ]

        if not isinstance(self.bottom_center_intensity_mean, list):
            self.bottom_center_intensity_mean = (
                [self.bottom_center_intensity_mean]
                if self.bottom_center_intensity_mean is not None
                else []
            )
        self.bottom_center_intensity_mean = [
            v if isinstance(v, float) else float(v) for v in self.bottom_center_intensity_mean
        ]

        if not isinstance(self.bottom_center_intensity_ratio, list):
            self.bottom_center_intensity_ratio = (
                [self.bottom_center_intensity_ratio]
                if self.bottom_center_intensity_ratio is not None
                else []
            )
        self.bottom_center_intensity_ratio = [
            v if isinstance(v, float) else float(v) for v in self.bottom_center_intensity_ratio
        ]

        if not isinstance(self.bottom_right_intensity_mean, list):
            self.bottom_right_intensity_mean = (
                [self.bottom_right_intensity_mean]
                if self.bottom_right_intensity_mean is not None
                else []
            )
        self.bottom_right_intensity_mean = [
            v if isinstance(v, float) else float(v) for v in self.bottom_right_intensity_mean
        ]

        if not isinstance(self.bottom_right_intensity_ratio, list):
            self.bottom_right_intensity_ratio = (
                [self.bottom_right_intensity_ratio]
                if self.bottom_right_intensity_ratio is not None
                else []
            )
        self.bottom_right_intensity_ratio = [
            v if isinstance(v, float) else float(v) for v in self.bottom_right_intensity_ratio
        ]

        if not isinstance(self.decile_0, list):
            self.decile_0 = [self.decile_0] if self.decile_0 is not None else []
        self.decile_0 = [v if isinstance(v, float) else float(v) for v in self.decile_0]

        if not isinstance(self.decile_1, list):
            self.decile_1 = [self.decile_1] if self.decile_1 is not None else []
        self.decile_1 = [v if isinstance(v, float) else float(v) for v in self.decile_1]

        if not isinstance(self.decile_2, list):
            self.decile_2 = [self.decile_2] if self.decile_2 is not None else []
        self.decile_2 = [v if isinstance(v, float) else float(v) for v in self.decile_2]

        if not isinstance(self.decile_3, list):
            self.decile_3 = [self.decile_3] if self.decile_3 is not None else []
        self.decile_3 = [v if isinstance(v, float) else float(v) for v in self.decile_3]

        if not isinstance(self.decile_4, list):
            self.decile_4 = [self.decile_4] if self.decile_4 is not None else []
        self.decile_4 = [v if isinstance(v, float) else float(v) for v in self.decile_4]

        if not isinstance(self.decile_5, list):
            self.decile_5 = [self.decile_5] if self.decile_5 is not None else []
        self.decile_5 = [v if isinstance(v, float) else float(v) for v in self.decile_5]

        if not isinstance(self.decile_6, list):
            self.decile_6 = [self.decile_6] if self.decile_6 is not None else []
        self.decile_6 = [v if isinstance(v, float) else float(v) for v in self.decile_6]

        if not isinstance(self.decile_7, list):
            self.decile_7 = [self.decile_7] if self.decile_7 is not None else []
        self.decile_7 = [v if isinstance(v, float) else float(v) for v in self.decile_7]

        if not isinstance(self.decile_8, list):
            self.decile_8 = [self.decile_8] if self.decile_8 is not None else []
        self.decile_8 = [v if isinstance(v, float) else float(v) for v in self.decile_8]

        if not isinstance(self.decile_9, list):
            self.decile_9 = [self.decile_9] if self.decile_9 is not None else []
        self.decile_9 = [v if isinstance(v, float) else float(v) for v in self.decile_9]

        super().__post_init__(**kwargs)


# Enumerations


# Slots
class slots:
    pass


slots.field_illumination_image = Slot(
    uri=DEFAULT_.field_illumination_image,
    name="field_illumination_image",
    curie=DEFAULT_.curie("field_illumination_image"),
    model_uri=DEFAULT_.field_illumination_image,
    domain=None,
    range=Union[dict, ImageAsNumpy],
)

slots.center_threshold = Slot(
    uri=DEFAULT_.center_threshold,
    name="center_threshold",
    curie=DEFAULT_.curie("center_threshold"),
    model_uri=DEFAULT_.center_threshold,
    domain=None,
    range=float,
)

slots.corner_fraction = Slot(
    uri=DEFAULT_.corner_fraction,
    name="corner_fraction",
    curie=DEFAULT_.curie("corner_fraction"),
    model_uri=DEFAULT_.corner_fraction,
    domain=None,
    range=float,
)

slots.sigma = Slot(
    uri=DEFAULT_.sigma,
    name="sigma",
    curie=DEFAULT_.curie("sigma"),
    model_uri=DEFAULT_.sigma,
    domain=None,
    range=float,
)

slots.intensity_map_size = Slot(
    uri=DEFAULT_.intensity_map_size,
    name="intensity_map_size",
    curie=DEFAULT_.curie("intensity_map_size"),
    model_uri=DEFAULT_.intensity_map_size,
    domain=None,
    range=int,
)

slots.channel = Slot(
    uri=DEFAULT_.channel,
    name="channel",
    curie=DEFAULT_.curie("channel"),
    model_uri=DEFAULT_.channel,
    domain=None,
    range=Optional[Union[int, List[int]]],
)

slots.nb_pixels = Slot(
    uri=DEFAULT_.nb_pixels,
    name="nb_pixels",
    curie=DEFAULT_.curie("nb_pixels"),
    model_uri=DEFAULT_.nb_pixels,
    domain=None,
    range=Optional[Union[int, List[int]]],
)

slots.center_of_mass_x = Slot(
    uri=DEFAULT_.center_of_mass_x,
    name="center_of_mass_x",
    curie=DEFAULT_.curie("center_of_mass_x"),
    model_uri=DEFAULT_.center_of_mass_x,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.center_of_mass_y = Slot(
    uri=DEFAULT_.center_of_mass_y,
    name="center_of_mass_y",
    curie=DEFAULT_.curie("center_of_mass_y"),
    model_uri=DEFAULT_.center_of_mass_y,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.max_intensity = Slot(
    uri=DEFAULT_.max_intensity,
    name="max_intensity",
    curie=DEFAULT_.curie("max_intensity"),
    model_uri=DEFAULT_.max_intensity,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.max_intensity_pos_x = Slot(
    uri=DEFAULT_.max_intensity_pos_x,
    name="max_intensity_pos_x",
    curie=DEFAULT_.curie("max_intensity_pos_x"),
    model_uri=DEFAULT_.max_intensity_pos_x,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.max_intensity_pos_y = Slot(
    uri=DEFAULT_.max_intensity_pos_y,
    name="max_intensity_pos_y",
    curie=DEFAULT_.curie("max_intensity_pos_y"),
    model_uri=DEFAULT_.max_intensity_pos_y,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.top_left_intensity_mean = Slot(
    uri=DEFAULT_.top_left_intensity_mean,
    name="top_left_intensity_mean",
    curie=DEFAULT_.curie("top_left_intensity_mean"),
    model_uri=DEFAULT_.top_left_intensity_mean,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.top_left_intensity_ratio = Slot(
    uri=DEFAULT_.top_left_intensity_ratio,
    name="top_left_intensity_ratio",
    curie=DEFAULT_.curie("top_left_intensity_ratio"),
    model_uri=DEFAULT_.top_left_intensity_ratio,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.top_center_intensity_mean = Slot(
    uri=DEFAULT_.top_center_intensity_mean,
    name="top_center_intensity_mean",
    curie=DEFAULT_.curie("top_center_intensity_mean"),
    model_uri=DEFAULT_.top_center_intensity_mean,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.top_center_intensity_ratio = Slot(
    uri=DEFAULT_.top_center_intensity_ratio,
    name="top_center_intensity_ratio",
    curie=DEFAULT_.curie("top_center_intensity_ratio"),
    model_uri=DEFAULT_.top_center_intensity_ratio,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.top_right_intensity_mean = Slot(
    uri=DEFAULT_.top_right_intensity_mean,
    name="top_right_intensity_mean",
    curie=DEFAULT_.curie("top_right_intensity_mean"),
    model_uri=DEFAULT_.top_right_intensity_mean,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.top_right_intensity_ratio = Slot(
    uri=DEFAULT_.top_right_intensity_ratio,
    name="top_right_intensity_ratio",
    curie=DEFAULT_.curie("top_right_intensity_ratio"),
    model_uri=DEFAULT_.top_right_intensity_ratio,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.middle_left_intensity_mean = Slot(
    uri=DEFAULT_.middle_left_intensity_mean,
    name="middle_left_intensity_mean",
    curie=DEFAULT_.curie("middle_left_intensity_mean"),
    model_uri=DEFAULT_.middle_left_intensity_mean,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.middle_left_intensity_ratio = Slot(
    uri=DEFAULT_.middle_left_intensity_ratio,
    name="middle_left_intensity_ratio",
    curie=DEFAULT_.curie("middle_left_intensity_ratio"),
    model_uri=DEFAULT_.middle_left_intensity_ratio,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.middle_center_intensity_mean = Slot(
    uri=DEFAULT_.middle_center_intensity_mean,
    name="middle_center_intensity_mean",
    curie=DEFAULT_.curie("middle_center_intensity_mean"),
    model_uri=DEFAULT_.middle_center_intensity_mean,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.middle_center_intensity_ratio = Slot(
    uri=DEFAULT_.middle_center_intensity_ratio,
    name="middle_center_intensity_ratio",
    curie=DEFAULT_.curie("middle_center_intensity_ratio"),
    model_uri=DEFAULT_.middle_center_intensity_ratio,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.middle_right_intensity_mean = Slot(
    uri=DEFAULT_.middle_right_intensity_mean,
    name="middle_right_intensity_mean",
    curie=DEFAULT_.curie("middle_right_intensity_mean"),
    model_uri=DEFAULT_.middle_right_intensity_mean,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.middle_right_intensity_ratio = Slot(
    uri=DEFAULT_.middle_right_intensity_ratio,
    name="middle_right_intensity_ratio",
    curie=DEFAULT_.curie("middle_right_intensity_ratio"),
    model_uri=DEFAULT_.middle_right_intensity_ratio,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.bottom_left_intensity_mean = Slot(
    uri=DEFAULT_.bottom_left_intensity_mean,
    name="bottom_left_intensity_mean",
    curie=DEFAULT_.curie("bottom_left_intensity_mean"),
    model_uri=DEFAULT_.bottom_left_intensity_mean,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.bottom_left_intensity_ratio = Slot(
    uri=DEFAULT_.bottom_left_intensity_ratio,
    name="bottom_left_intensity_ratio",
    curie=DEFAULT_.curie("bottom_left_intensity_ratio"),
    model_uri=DEFAULT_.bottom_left_intensity_ratio,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.bottom_center_intensity_mean = Slot(
    uri=DEFAULT_.bottom_center_intensity_mean,
    name="bottom_center_intensity_mean",
    curie=DEFAULT_.curie("bottom_center_intensity_mean"),
    model_uri=DEFAULT_.bottom_center_intensity_mean,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.bottom_center_intensity_ratio = Slot(
    uri=DEFAULT_.bottom_center_intensity_ratio,
    name="bottom_center_intensity_ratio",
    curie=DEFAULT_.curie("bottom_center_intensity_ratio"),
    model_uri=DEFAULT_.bottom_center_intensity_ratio,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.bottom_right_intensity_mean = Slot(
    uri=DEFAULT_.bottom_right_intensity_mean,
    name="bottom_right_intensity_mean",
    curie=DEFAULT_.curie("bottom_right_intensity_mean"),
    model_uri=DEFAULT_.bottom_right_intensity_mean,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.bottom_right_intensity_ratio = Slot(
    uri=DEFAULT_.bottom_right_intensity_ratio,
    name="bottom_right_intensity_ratio",
    curie=DEFAULT_.curie("bottom_right_intensity_ratio"),
    model_uri=DEFAULT_.bottom_right_intensity_ratio,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.decile_0 = Slot(
    uri=DEFAULT_.decile_0,
    name="decile_0",
    curie=DEFAULT_.curie("decile_0"),
    model_uri=DEFAULT_.decile_0,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.decile_1 = Slot(
    uri=DEFAULT_.decile_1,
    name="decile_1",
    curie=DEFAULT_.curie("decile_1"),
    model_uri=DEFAULT_.decile_1,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.decile_2 = Slot(
    uri=DEFAULT_.decile_2,
    name="decile_2",
    curie=DEFAULT_.curie("decile_2"),
    model_uri=DEFAULT_.decile_2,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.decile_3 = Slot(
    uri=DEFAULT_.decile_3,
    name="decile_3",
    curie=DEFAULT_.curie("decile_3"),
    model_uri=DEFAULT_.decile_3,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.decile_4 = Slot(
    uri=DEFAULT_.decile_4,
    name="decile_4",
    curie=DEFAULT_.curie("decile_4"),
    model_uri=DEFAULT_.decile_4,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.decile_5 = Slot(
    uri=DEFAULT_.decile_5,
    name="decile_5",
    curie=DEFAULT_.curie("decile_5"),
    model_uri=DEFAULT_.decile_5,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.decile_6 = Slot(
    uri=DEFAULT_.decile_6,
    name="decile_6",
    curie=DEFAULT_.curie("decile_6"),
    model_uri=DEFAULT_.decile_6,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.decile_7 = Slot(
    uri=DEFAULT_.decile_7,
    name="decile_7",
    curie=DEFAULT_.curie("decile_7"),
    model_uri=DEFAULT_.decile_7,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.decile_8 = Slot(
    uri=DEFAULT_.decile_8,
    name="decile_8",
    curie=DEFAULT_.curie("decile_8"),
    model_uri=DEFAULT_.decile_8,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.decile_9 = Slot(
    uri=DEFAULT_.decile_9,
    name="decile_9",
    curie=DEFAULT_.curie("decile_9"),
    model_uri=DEFAULT_.decile_9,
    domain=None,
    range=Optional[Union[float, List[float]]],
)

slots.fieldIlluminationDataset__input = Slot(
    uri=DEFAULT_.input,
    name="fieldIlluminationDataset__input",
    curie=DEFAULT_.curie("input"),
    model_uri=DEFAULT_.fieldIlluminationDataset__input,
    domain=None,
    range=Union[dict, FieldIlluminationInput],
)

slots.fieldIlluminationDataset__output = Slot(
    uri=DEFAULT_.output,
    name="fieldIlluminationDataset__output",
    curie=DEFAULT_.curie("output"),
    model_uri=DEFAULT_.fieldIlluminationDataset__output,
    domain=None,
    range=Optional[Union[dict, FieldIlluminationOutput]],
)

slots.fieldIlluminationOutput__key_values = Slot(
    uri=DEFAULT_.key_values,
    name="fieldIlluminationOutput__key_values",
    curie=DEFAULT_.curie("key_values"),
    model_uri=DEFAULT_.fieldIlluminationOutput__key_values,
    domain=None,
    range=Optional[Union[dict, FieldIlluminationKeyValues]],
)

slots.fieldIlluminationOutput__intensity_profiles = Slot(
    uri=DEFAULT_.intensity_profiles,
    name="fieldIlluminationOutput__intensity_profiles",
    curie=DEFAULT_.curie("intensity_profiles"),
    model_uri=DEFAULT_.fieldIlluminationOutput__intensity_profiles,
    domain=None,
    range=Optional[Union[dict, TableAsDict]],
)

slots.fieldIlluminationOutput__intensity_map = Slot(
    uri=DEFAULT_.intensity_map,
    name="fieldIlluminationOutput__intensity_map",
    curie=DEFAULT_.curie("intensity_map"),
    model_uri=DEFAULT_.fieldIlluminationOutput__intensity_map,
    domain=None,
    range=Optional[Union[str, Image5DImageUrl]],
)

slots.fieldIlluminationOutput__profile_rois = Slot(
    uri=DEFAULT_.profile_rois,
    name="fieldIlluminationOutput__profile_rois",
    curie=DEFAULT_.curie("profile_rois"),
    model_uri=DEFAULT_.fieldIlluminationOutput__profile_rois,
    domain=None,
    range=Optional[Union[dict, ROI]],
)



# Main samples module defining the sample superclass

import logging
from abc import ABC, abstractmethod
from typing import List, Union

import numpy as np

from microscopemetrics.data_schema import core_schema

# We are defining some global dictionaries to register the different analysis types
IMAGE_ANALYSIS_REGISTRY = {}
DATASET_ANALYSIS_REGISTRY = {}
PROGRESSION_ANALYSIS_REGISTRY = {}


# Decorators to register exposed analysis functions
def register_image_analysis(cls):
    IMAGE_ANALYSIS_REGISTRY[cls.__name__] = cls
    return cls


def register_dataset_analysis(cls):
    DATASET_ANALYSIS_REGISTRY[cls.__name__] = cls
    return cls


def register_progression_analysis(cls):
    PROGRESSION_ANALYSIS_REGISTRY[cls.__name__] = cls
    return cls


# Create a logging service
logger = logging.getLogger(__name__)


def numpy_to_inlined_mask(
    array: np.ndarray,
    name: str = None,
    description: str = None,
    image_url: str = None,
    source_image_url: Union[str, List[str]] = None,
) -> core_schema.ImageMask:
    """Converts a bool numpy array to an inlined mask"""
    if array.ndim != 2:
        raise ValueError("Input array should be 2D")
    if array.dtype != bool and array.dtype == np.uint8:
        try:
            array = array.astype(bool)
        except ValueError:
            raise ValueError("Input array could not be casted to type bool")
    if array.dtype != bool:
        raise ValueError("Input array should be of type bool")

    return core_schema.ImageMask(
        name=name,
        description=description,
        image_url=image_url,
        source_image_url=source_image_url,
        data=array.flatten().tolist(),
        y=core_schema.PixelSeries(values=array.shape[0]),
        x=core_schema.PixelSeries(values=array.shape[1]),
    )


def numpy_to_inlined_image(
    array: np.ndarray,
    name: str = None,
    description: str = None,
    image_url: str = None,
    source_image_url: Union[str, List[str]] = None,
) -> core_schema.ImageInline:
    """Converts a numpy array to an inlined image"""
    if array.ndim == 5:
        return core_schema.Image5D(
            name=name,
            description=description,
            image_url=image_url,
            source_image_url=source_image_url,
            data=array.flatten().tolist(),
            t=core_schema.TimeSeries(values=array.shape[0]),
            z=core_schema.PixelSeries(values=array.shape[1]),
            y=core_schema.PixelSeries(values=array.shape[2]),
            x=core_schema.PixelSeries(values=array.shape[3]),
            c=core_schema.ChannelSeries(values=array.shape[4]),
        )
    elif array.ndim == 2:
        return core_schema.Image2D(
            name=name,
            description=description,
            image_url=image_url,
            source_image_url=source_image_url,
            data=array.flatten().tolist(),
            y=core_schema.PixelSeries(values=array.shape[0]),
            x=core_schema.PixelSeries(values=array.shape[1]),
        )
    else:
        raise NotImplementedError(
            f"Array of dimension {array.ndim} is not supported by this function"
        )


def dict_to_inlined_table(
    dictionary: dict[str, list],
    name: str = None,
    description: str = None,
) -> core_schema.Table:
    """Converts a dictionary to an microscope-metrics inlined table"""
    return core_schema.TableAsDict(
        name=name,
        description=description,
        columns=[dict(zip(dictionary, row)) for row in zip(*dictionary.values())],
    )


class AnalysisMixin(ABC):
    """This is a mixin class defining some helper functions to work with the linkml autogenerated schema."""

    def __init__(self):
        pass

    def describe_requirements(self):
        print(self.input.describe_requirements())

    def validate_requirements(self) -> bool:
        logger.info("Validating requirements...")
        # TODO: check image dimensions/shape
        return True
